# === Module import and public API surface ===
import binascii

expected_public = [
    'Error',
    'Incomplete',
    'a2b_base64',
    'a2b_hex',
    'a2b_qp',
    'a2b_uu',
    'b2a_base64',
    'b2a_hex',
    'b2a_qp',
    'b2a_uu',
    'crc32',
    'crc_hqx',
    'hexlify',
    'unhexlify',
]
for name in expected_public:
    assert hasattr(binascii, name), f'missing public attribute: {name}'


# === Exception classes ===
err = binascii.Error('boom')
inc = binascii.Incomplete('partial')
assert isinstance(err, Exception), 'Error should be an exception type'
assert isinstance(inc, Exception), 'Incomplete should be an exception type'
assert err.args == ('boom',), 'Error args should be preserved'
assert inc.args == ('partial',), 'Incomplete args should be preserved'


def expect_error(fn, expected_message):
    try:
        fn()
        assert False, f'expected error: {expected_message}'
    except Exception as exc:
        assert str(exc) == expected_message, f'expected {expected_message!r}, got {str(exc)!r}'


# === hexlify / b2a_hex ===
assert binascii.hexlify(b'\x00\xffab') == b'00ff6162', 'hexlify basic bytes'
assert binascii.hexlify(bytearray(b'\x00\xffab')) == b'00ff6162', 'hexlify accepts bytearray'
assert binascii.hexlify(b'UUDDLRLRAB', b':', 2) == b'5555:4444:4c52:4c52:4142', 'hexlify sep from right'
assert binascii.hexlify(b'UUDDLRLRAB', ':', -4) == b'55554444:4c524c52:4142', 'hexlify sep from left'
assert binascii.hexlify(b'abc', bytes_per_sep=2) == b'616263', 'bytes_per_sep ignored without sep'
assert binascii.b2a_hex(b'abc', b':', 2) == b'61:6263', 'b2a_hex alias behavior'

expect_error(lambda: binascii.hexlify(b'abc', b'--'), 'sep must be length 1.')
expect_error(lambda: binascii.hexlify('abc'), "a bytes-like object is required, not 'str'")


# === unhexlify / a2b_hex ===
assert binascii.unhexlify(b'00ff6162') == b'\x00\xffab', 'unhexlify bytes input'
assert binascii.unhexlify('00ff6162') == b'\x00\xffab', 'unhexlify str input'
assert binascii.unhexlify(bytearray(b'616263')) == b'abc', 'unhexlify accepts bytearray'
assert binascii.a2b_hex('6162') == b'ab', 'a2b_hex alias behavior'

expect_error(lambda: binascii.unhexlify('abc'), 'Odd-length string')
expect_error(lambda: binascii.unhexlify('abxz'), 'Non-hexadecimal digit found')
expect_error(lambda: binascii.unhexlify('abé0'), 'string argument should contain only ASCII characters')
expect_error(lambda: binascii.unhexlify(1), "argument should be bytes, buffer or ASCII string, not 'int'")


# === crc32 / crc_hqx ===
assert binascii.crc32(b'hello') == 907060870, 'crc32 basic'
assert binascii.crc32(b'hello', 1) == 191926070, 'crc32 with seed'
assert binascii.crc32(b'hello', -1) == 265137764, 'crc32 wraps negative seed'
assert binascii.crc32(b'hello', 2**65 + 3) == 1907416150, 'crc32 wraps large seed'

assert binascii.crc_hqx(b'123456789', 0) == 12739, 'crc_hqx basic vector'
assert binascii.crc_hqx(b'123456789', 0xFFFF) == 10673, 'crc_hqx with nonzero seed'
assert binascii.crc_hqx(b'abc', -1) == 20810, 'crc_hqx wraps negative seed'
assert binascii.crc_hqx(b'abc', 1 << 40) == 40406, 'crc_hqx wraps large seed'

expect_error(lambda: binascii.crc32('hello'), "a bytes-like object is required, not 'str'")
expect_error(lambda: binascii.crc_hqx('hello', 0), "a bytes-like object is required, not 'str'")


# === a2b_base64 / b2a_base64 ===
assert binascii.a2b_base64(b'YQ==') == b'a', 'a2b_base64 bytes input'
assert binascii.a2b_base64('YQ==') == b'a', 'a2b_base64 str input'
assert binascii.a2b_base64(b'Y!Q==') == b'a', 'a2b_base64 non-strict ignores non-base64 bytes'
assert binascii.a2b_base64(b'YQ==', strict_mode=True) == b'a', 'a2b_base64 strict valid input'
expect_error(lambda: binascii.a2b_base64(b'Y Q==', strict_mode=True), 'Only base64 data is allowed')
expect_error(
    lambda: binascii.a2b_base64(b'A'),
    'Invalid base64-encoded string: number of data characters (1) cannot be 1 more than a multiple of 4',
)

assert binascii.b2a_base64(b'a') == b'YQ==\n', 'b2a_base64 default newline'
assert binascii.b2a_base64(b'a', newline=False) == b'YQ==', 'b2a_base64 newline=False'


# === a2b_qp / b2a_qp ===
assert binascii.a2b_qp(b'a=3Db') == b'a=b', 'a2b_qp hex decode'
assert binascii.a2b_qp(b'a_b', header=True) == b'a b', 'a2b_qp header underscore-to-space'
assert binascii.a2b_qp(b'abc=\nxyz') == b'abcxyz', 'a2b_qp soft line break'
assert binascii.a2b_qp('a=3Db') == b'a=b', 'a2b_qp ascii string input'
expect_error(lambda: binascii.a2b_qp('é'), 'string argument should contain only ASCII characters')

assert binascii.b2a_qp(b'a=b') == b'a=3Db', 'b2a_qp equals escaping'
assert binascii.b2a_qp(b'a\tb', quotetabs=True) == b'a=09b', 'b2a_qp quotetabs'
assert binascii.b2a_qp(b'a b', header=True) == b'a_b', 'b2a_qp header space-to-underscore'
assert binascii.b2a_qp(b'a\n', istext=False) == b'a=0A', 'b2a_qp istext=False newline escaping'
expect_error(lambda: binascii.b2a_qp('a=b'), "a bytes-like object is required, not 'str'")


# === a2b_uu / b2a_uu ===
assert binascii.b2a_uu(b'cat') == b'#8V%T\n', 'b2a_uu basic'
assert binascii.b2a_uu(b'\x00', backtick=True) == b'!````\n', 'b2a_uu backtick'
expect_error(lambda: binascii.b2a_uu(b'a' * 46), 'At most 45 bytes at once')

assert binascii.a2b_uu(b'#0V%T\n') == b'Cat', 'a2b_uu basic decode'
assert binascii.a2b_uu(b'#0V') == b'C`\x00', 'a2b_uu short input zero-pads'
assert binascii.a2b_uu('') == (b'\x00' * 32), 'a2b_uu empty string compatibility behavior'
expect_error(lambda: binascii.a2b_uu(b'!\xff\n'), 'Illegal char')
expect_error(lambda: binascii.a2b_uu(b'#0V%Txx'), 'Trailing garbage')



# === Basic function calls ===
def f_no_args():
    return 1


assert f_no_args() == 1, 'no args'


def f_one_arg(x):
    return x


assert f_one_arg(42) == 42, 'one arg'


def add(a, b):
    return a + b


assert add(1, 2) == 3, 'two args'


def sum3(a, b, c):
    return a + b + c


assert sum3(1, 2, 3) == 6, 'three args'


# === Local variables ===
def f_local():
    x = 42
    return x


assert f_local() == 42, 'local var'


def f_local_from_arg(x):
    y = x + 1
    return y


assert f_local_from_arg(10) == 11, 'local var from arg'


def f_local_list():
    items = [1, 2, 3]
    return items


assert f_local_list() == [1, 2, 3], 'local var list'


def f_local_modify_list():
    items = [1, 2]
    items.append(3)
    return items


assert f_local_modify_list() == [1, 2, 3], 'local var modify list'


def f_local_multiple():
    a = 1
    b = 2
    c = 3
    return a + b + c


assert f_local_multiple() == 6, 'local var multiple'


def f_local_reassign():
    x = 1
    x = 2
    x = 3
    return x


assert f_local_reassign() == 3, 'local var reassign'


# === Nested functions ===
def nested_basic():
    def bar():
        return 1

    return bar() + 1


assert nested_basic() == 2, 'nested basic'


def nested_deep():
    def level2():
        def level3():
            return 42

        return level3()

    return level2()


assert nested_deep() == 42, 'nested deep'


def nested_multiple_calls():
    def inner():
        return 10

    return inner() + inner() + inner()


assert nested_multiple_calls() == 30, 'nested multiple calls'


def nested_two_inner():
    def add():
        return 1

    def sub():
        return 2

    return add() + sub()


assert nested_two_inner() == 3, 'nested two inner'


def nested_with_args(x):
    def inner(y):
        return y + y

    return inner(x) + 1


assert nested_with_args(5) == 11, 'nested with args'


# === Function equality ===
def eq_test():
    return 1


def eq_test2():
    return 1


# Same function is equal to itself
assert eq_test == eq_test, 'function equals itself'
assert not (eq_test != eq_test), 'function not-not-equals itself'

# Different functions are not equal (even with same body)
assert not (eq_test == eq_test2), 'different functions not equal'
assert eq_test != eq_test2, 'different functions are not equal'

# Function assigned to variable is still equal
f_alias = eq_test
assert f_alias == eq_test, 'function alias equals original'
assert eq_test == f_alias, 'original equals function alias'


# === Builtin equality ===
# Same builtin is equal to itself
assert len == len, 'builtin equals itself'
assert print == print, 'print equals itself'
assert not (len != len), 'builtin not-not-equals itself'

# Builtin identity (is)
assert print is print, 'print is print'
assert len is len, 'len is len'
assert not (len is print), 'len is not print'

# Different builtins are not equal
assert not (len == print), 'different builtins not equal'
assert len != print, 'different builtins are not equal'

# Builtin assigned to variable is still equal
len_alias = len
assert len_alias == len, 'builtin alias equals original'
assert len_alias is len, 'builtin alias is original'


# === Exception type equality ===
# Note: Using == instead of 'is' to explicitly test the __eq__ implementation
assert ValueError == ValueError, 'exc type equals itself'
assert TypeError == TypeError, 'exc type equals itself 2'
assert not (ValueError != ValueError), 'exc type not-not-equals itself'

assert not (ValueError == TypeError), 'different exc types not equal'
assert ValueError != TypeError, 'different exc types are not equal'

exc_alias = ValueError
assert exc_alias == ValueError, 'exc type alias equals original'


# === Closure equality ===
def make_adder(n):
    def adder(x):
        return x + n

    return adder


add1 = make_adder(1)
add2 = make_adder(2)
add1_again = make_adder(1)

# Same closure instance equals itself
assert add1 == add1, 'closure equals itself'
assert not (add1 != add1), 'closure not-not-equals itself'

# Different closure instances are not equal (even with same captured value)
assert not (add1 == add1_again), 'different closure instances not equal'
assert add1 != add1_again, 'different closure instances are not equal'

# Different closure instances with different captured values
assert not (add1 == add2), 'closures with diff captured values not equal'
assert add1 != add2, 'closures with diff captured values are not equal'


# === Cross-type inequality ===
def cross_test():
    return 1


assert not (cross_test == len), 'function not equal to builtin'
assert not (len == cross_test), 'builtin not equal to function'
assert not (cross_test == ValueError), 'function not equal to exc type'
assert not (ValueError == cross_test), 'exc type not equal to function'
assert not (len == ValueError), 'builtin not equal to exc type'
assert not (ValueError == len), 'exc type not equal to builtin'

# Callables not equal to other types
assert not (len == 1), 'builtin not equal to int'
assert not (len == 'len'), 'builtin not equal to string'
assert not (cross_test == None), 'function not equal to None'
assert not (ValueError == None), 'exc type not equal to None'


# === Parameter shadowing global variables ===
# Function parameters should shadow global variables with the same name
x = 5


def shadow_single(x):
    return x + 1


# When called with 10, param x=10 should be used, not global x=5
assert shadow_single(10) == 11, 'param shadows global - single param'

y = 3


def shadow_multiple(x, y):
    return x + y


# When called with (20, 30), params should be used, not globals x=5, y=3
assert shadow_multiple(20, 30) == 50, 'param shadows global - multiple params'


def shadow_uses_global_too(x):
    # x is param, y is global
    return x + y


# x=100 (param), y=3 (global), so 100 + 3 = 103
assert shadow_uses_global_too(100) == 103, 'param shadows but can still access other globals'


def shadow_with_default(x=99):
    return x + 1


# When called with argument, param shadows global
assert shadow_with_default(10) == 11, 'param with default shadows global'
# When called without argument, default is used (not global)
assert shadow_with_default() == 100, 'param default used, not global'


# Global is still accessible outside the function
assert x == 5, 'global still accessible after function that shadows it'
assert y == 3, 'other global still accessible'


# Verify global can still be used as argument
def double(x):
    return x * 2


assert double(x) == 10, 'global used as argument, param shadows inside'


# === Builtin name shadowing in locals/parameters ===
def shadow_builtin_params(id, len, type, max):
    return id, len, type, max


assert shadow_builtin_params('hello', 42, 'custom', 99) == (
    'hello',
    42,
    'custom',
    99,
), 'builtin-name params resolve to local values'


def shadow_builtin_locals():
    id = 'local-id'
    len = 7
    type = 'local-type'
    max = 'local-max'
    return id, len, type, max


assert shadow_builtin_locals() == (
    'local-id',
    7,
    'local-type',
    'local-max',
), 'builtin-name locals resolve to local values'


# === Function repr formatting ===
def repr_target():
    return 123


def repr_outer():
    def repr_inner():
        return 456

    return repr_inner


repr_inner_fn = repr_outer()
assert repr(repr_target).startswith('<function repr_target at 0x'), 'top-level function repr format'
assert repr(repr_inner_fn).startswith('<function repr_outer.<locals>.repr_inner at 0x'), 'nested function repr format'



# === Break in try/finally must run finally ===
result = []
for x in [1, 2, 3]:
    try:
        result.append('before')
        break
        result.append('after')  # unreachable
    finally:
        result.append('finally')
assert result == ['before', 'finally'], f'break in try/finally should run finally: {result}'

# === Break in nested try/finally runs both finally blocks ===
result = []
for x in [1, 2, 3]:
    try:
        try:
            result.append('inner-try')
            break
        finally:
            result.append('inner-finally')
    finally:
        result.append('outer-finally')
assert result == ['inner-try', 'inner-finally', 'outer-finally'], f'nested finally blocks: {result}'

# === Break in try/except/finally runs finally ===
result = []
for x in [1, 2, 3]:
    try:
        result.append('try')
        break
    except ValueError:
        result.append('except')
    finally:
        result.append('finally')
assert result == ['try', 'finally'], f'break in try/except/finally: {result}'

# === Break inside except handler with finally ===
result = []
for x in [1, 2, 3]:
    try:
        raise ValueError('test')
    except ValueError:
        result.append('except')
        break
    finally:
        result.append('finally')
assert result == ['except', 'finally'], f'break in except with finally: {result}'

# === Break does not run finally if not in try ===
result = []
for x in [1, 2, 3]:
    result.append('body')
    break
assert result == ['body'], f'break without finally: {result}'

# === Break with multiple loops and finally ===
result = []
for i in [1, 2]:
    try:
        for j in [10, 20, 30]:
            if j == 20:
                break  # This break should not trigger outer finally
            result.append(j)
        result.append('after-inner')
    finally:
        result.append('outer-finally')
assert result == [10, 'after-inner', 'outer-finally', 10, 'after-inner', 'outer-finally'], (
    f'inner break with outer finally: {result}'
)



import gc

# === collect / enabled state ===
assert isinstance(gc.collect(), int), 'collect_returns_int'
assert gc.disable() is None, 'disable_returns_none'
assert gc.isenabled() is False, 'isenabled_after_disable'
assert gc.enable() is None, 'enable_returns_none'
assert gc.isenabled() is True, 'isenabled_after_enable'

# === thresholds ===
thresholds = gc.get_threshold()
assert isinstance(thresholds, tuple), 'thresholds_type'
assert len(thresholds) == 3, 'thresholds_len'
assert all(isinstance(v, int) for v in thresholds), 'thresholds_ints'
assert gc.set_threshold(*thresholds) is None, 'set_threshold_roundtrip'
assert gc.get_threshold() == thresholds, 'thresholds_roundtrip'

# === debug flags ===
debug_before = gc.get_debug()
assert isinstance(debug_before, int), 'get_debug_type'
assert gc.set_debug(debug_before | 1) is None, 'set_debug_returns_none'
assert isinstance(gc.get_debug(), int), 'get_debug_after_set_type'
assert gc.set_debug(debug_before) is None, 'set_debug_restore'

# === stats and counts ===
counts = gc.get_count()
assert isinstance(counts, tuple), 'get_count_type'
assert len(counts) == 3, 'get_count_len'
assert all(isinstance(v, int) for v in counts), 'get_count_values'

stats = gc.get_stats()
assert isinstance(stats, list), 'get_stats_type'
for row in stats:
    assert isinstance(row, dict), 'get_stats_row_type'
    assert set(row.keys()) == {'collections', 'collected', 'uncollectable'}, 'get_stats_row_keys'

# === referents/referrers and object predicates ===
assert gc.get_referents() == [], 'get_referents_empty'
assert gc.get_referrers() == [], 'get_referrers_empty'
assert isinstance(gc.is_tracked([]), bool), 'is_tracked_bool'
assert isinstance(gc.is_finalized(object()), bool), 'is_finalized_bool'

# === freeze compatibility hooks ===
assert gc.freeze() is None, 'freeze_returns_none'
assert isinstance(gc.get_freeze_count(), int), 'get_freeze_count_type'
assert gc.unfreeze() is None, 'unfreeze_returns_none'



import _thread
import sys
from _thread import _excepthook, _ExceptHookArgs, get_native_id as get_native_id
from _typeshed import ProfileFunction, TraceFunction
from collections.abc import Callable, Iterable, Mapping
from contextvars import ContextVar
from types import TracebackType
from typing import Any, Final, TypeVar, final
from typing_extensions import deprecated

_T = TypeVar("_T")

__all__ = [
    "get_ident",
    "active_count",
    "Condition",
    "current_thread",
    "enumerate",
    "main_thread",
    "TIMEOUT_MAX",
    "Event",
    "Lock",
    "RLock",
    "Semaphore",
    "BoundedSemaphore",
    "Thread",
    "Barrier",
    "BrokenBarrierError",
    "Timer",
    "ThreadError",
    "ExceptHookArgs",
    "setprofile",
    "settrace",
    "local",
    "stack_size",
    "excepthook",
    "get_native_id",
]

if sys.version_info >= (3, 10):
    __all__ += ["getprofile", "gettrace"]

if sys.version_info >= (3, 12):
    __all__ += ["setprofile_all_threads", "settrace_all_threads"]

_profile_hook: ProfileFunction | None

def active_count() -> int: ...
@deprecated("Deprecated since Python 3.10. Use `active_count()` instead.")
def activeCount() -> int: ...
def current_thread() -> Thread: ...
@deprecated("Deprecated since Python 3.10. Use `current_thread()` instead.")
def currentThread() -> Thread: ...
def get_ident() -> int:
    """
    Return a non-zero integer that uniquely identifies the current thread
    amongst other threads that exist simultaneously.
    This may be used to identify per-thread resources.
    Even though on some platforms threads identities may appear to be
    allocated consecutive numbers starting at 1, this behavior should not
    be relied upon, and the number should be seen purely as a magic cookie.
    A thread's identity may be reused for another thread after it exits.
    """
    ...
def enumerate() -> list[Thread]: ...
def main_thread() -> Thread: ...
def settrace(func: TraceFunction) -> None: ...
def setprofile(func: ProfileFunction | None) -> None: ...

if sys.version_info >= (3, 12):
    def setprofile_all_threads(func: ProfileFunction | None) -> None: ...
    def settrace_all_threads(func: TraceFunction) -> None: ...

if sys.version_info >= (3, 10):
    def gettrace() -> TraceFunction | None: ...
    def getprofile() -> ProfileFunction | None: ...

def stack_size(size: int = 0, /) -> int:
    """
    Return the thread stack size used when creating new threads.  The
    optional size argument specifies the stack size (in bytes) to be used
    for subsequently created threads, and must be 0 (use platform or
    configured default) or a positive integer value of at least 32,768 (32k).
    If changing the thread stack size is unsupported, a ThreadError
    exception is raised.  If the specified size is invalid, a ValueError
    exception is raised, and the stack size is unmodified.  32k bytes
     currently the minimum supported stack size value to guarantee
    sufficient stack space for the interpreter itself.

    Note that some platforms may have particular restrictions on values for
    the stack size, such as requiring a minimum stack size larger than 32 KiB or
    requiring allocation in multiples of the system memory page size
    - platform documentation should be referred to for more information
    (4 KiB pages are common; using multiples of 4096 for the stack size is
    the suggested approach in the absence of more specific information).
    """
    ...

TIMEOUT_MAX: Final[float]

ThreadError = _thread.error
local = _thread._local

class Thread:
    name: str
    @property
    def ident(self) -> int | None:
        """
        Thread identifier of this thread or None if it has not been started.

        This is a nonzero integer. See the get_ident() function. Thread
        identifiers may be recycled when a thread exits and another thread is
        created. The identifier is available even after the thread has exited.
        """
        ...
    daemon: bool
    if sys.version_info >= (3, 14):
        def __init__(
            self,
            group: None = None,
            target: Callable[..., object] | None = None,
            name: str | None = None,
            args: Iterable[Any] = (),
            kwargs: Mapping[str, Any] | None = None,
            *,
            daemon: bool | None = None,
            context: ContextVar[Any] | None = None,
        ) -> None: ...
    else:
        def __init__(
            self,
            group: None = None,
            target: Callable[..., object] | None = None,
            name: str | None = None,
            args: Iterable[Any] = (),
            kwargs: Mapping[str, Any] | None = None,
            *,
            daemon: bool | None = None,
        ) -> None: ...

    def start(self) -> None: ...
    def run(self) -> None: ...
    def join(self, timeout: float | None = None) -> None: ...
    @property
    def native_id(self) -> int | None:
        """
        Native integral thread ID of this thread, or None if it has not been started.

        This is a non-negative integer. See the get_native_id() function.
        This represents the Thread ID as reported by the kernel.
        """
        ...
    def is_alive(self) -> bool: ...
    @deprecated("Deprecated since Python 3.10. Read the `daemon` attribute instead.")
    def isDaemon(self) -> bool: ...
    @deprecated("Deprecated since Python 3.10. Set the `daemon` attribute instead.")
    def setDaemon(self, daemonic: bool) -> None: ...
    @deprecated("Deprecated since Python 3.10. Read the `name` attribute instead.")
    def getName(self) -> str: ...
    @deprecated("Deprecated since Python 3.10. Set the `name` attribute instead.")
    def setName(self, name: str) -> None: ...

class _DummyThread(Thread):
    def __init__(self) -> None: ...

# This is actually the function _thread.allocate_lock for <= 3.12
Lock = _thread.LockType

# Python implementation of RLock.
@final
class _RLock:
    _count: int
    def acquire(self, blocking: bool = True, timeout: float = -1) -> bool: ...
    def release(self) -> None: ...
    __enter__ = acquire
    def __exit__(self, t: type[BaseException] | None, v: BaseException | None, tb: TracebackType | None) -> None: ...

    if sys.version_info >= (3, 14):
        def locked(self) -> bool: ...

RLock = _thread.RLock  # Actually a function at runtime.

class Condition:
    def __init__(self, lock: Lock | _RLock | RLock | None = None) -> None: ...
    def __enter__(self) -> bool: ...
    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None: ...
    def acquire(self, blocking: bool = True, timeout: float = -1) -> bool: ...
    def release(self) -> None: ...
    def wait(self, timeout: float | None = None) -> bool: ...
    def wait_for(self, predicate: Callable[[], _T], timeout: float | None = None) -> _T: ...
    def notify(self, n: int = 1) -> None: ...
    def notify_all(self) -> None: ...
    @deprecated("Deprecated since Python 3.10. Use `notify_all()` instead.")
    def notifyAll(self) -> None: ...

class Semaphore:
    _value: int
    def __init__(self, value: int = 1) -> None: ...
    def __exit__(self, t: type[BaseException] | None, v: BaseException | None, tb: TracebackType | None) -> None: ...
    def acquire(self, blocking: bool = True, timeout: float | None = None) -> bool: ...
    def __enter__(self, blocking: bool = True, timeout: float | None = None) -> bool: ...
    def release(self, n: int = 1) -> None: ...

class BoundedSemaphore(Semaphore): ...

class Event:
    def is_set(self) -> bool: ...
    @deprecated("Deprecated since Python 3.10. Use `is_set()` instead.")
    def isSet(self) -> bool: ...
    def set(self) -> None: ...
    def clear(self) -> None: ...
    def wait(self, timeout: float | None = None) -> bool: ...

excepthook = _excepthook
ExceptHookArgs = _ExceptHookArgs

class Timer(Thread):
    args: Iterable[Any]  # undocumented
    finished: Event  # undocumented
    function: Callable[..., Any]  # undocumented
    interval: float  # undocumented
    kwargs: Mapping[str, Any]  # undocumented

    def __init__(
        self,
        interval: float,
        function: Callable[..., object],
        args: Iterable[Any] | None = None,
        kwargs: Mapping[str, Any] | None = None,
    ) -> None: ...
    def cancel(self) -> None: ...

class Barrier:
    @property
    def parties(self) -> int:
        """Return the number of threads required to trip the barrier."""
        ...
    @property
    def n_waiting(self) -> int:
        """Return the number of threads currently waiting at the barrier."""
        ...
    @property
    def broken(self) -> bool:
        """Return True if the barrier is in a broken state."""
        ...
    def __init__(self, parties: int, action: Callable[[], None] | None = None, timeout: float | None = None) -> None: ...
    def wait(self, timeout: float | None = None) -> int: ...
    def reset(self) -> None: ...
    def abort(self) -> None: ...

class BrokenBarrierError(RuntimeError): ...



# === Literal matching ===
def describe(x):
    match x:
        case 1:
            return 'one'
        case 2:
            return 'two'
        case _:
            return 'other'

assert describe(1) == 'one', 'literal match 1'
assert describe(2) == 'two', 'literal match 2'
assert describe(99) == 'other', 'wildcard match'

# === Variable capture ===
def get_first(point):
    match point:
        case (x, y):
            return x

assert get_first((10, 20)) == 10, 'tuple capture'

# === Guard clause ===
def classify(x):
    match x:
        case n if n < 0:
            return 'negative'
        case 0:
            return 'zero'
        case n if n > 0:
            return 'positive'

assert classify(-5) == 'negative', 'guard negative'
assert classify(0) == 'zero', 'guard zero'
assert classify(5) == 'positive', 'guard positive'

# === Or pattern ===
def weekend(day):
    match day:
        case 'Saturday' | 'Sunday':
            return True
        case _:
            return False

assert weekend('Saturday') == True, 'or pattern match'
assert weekend('Monday') == False, 'or pattern no match'
print('match/case tests passed')



# run-async
async def foo():
    return 1


coro = foo()
await coro  # pyright: ignore
await coro  # pyright: ignore
"""
TRACEBACK:
Traceback (most recent call last):
  File "async__double_await_coroutine.py", line 8, in <module>
    await coro  # pyright: ignore
    ~~~~~~~~~~
RuntimeError: cannot reuse already awaited coroutine
"""



# When breaking from nested except handlers, ALL exception states must be cleared.
# After the loop completes via break, execution should continue normally.

# Test 1: break from depth 2 should reach code after loop
def test_break():
    for i in range(1):
        try:
            raise ValueError('outer')
        except:
            try:
                raise TypeError('inner')
            except:
                break  # Should clear BOTH exceptions
    return 'ok'


assert test_break() == 'ok', 'break from nested except should reach return'


# Test 2: break from depth 3 should also work
def test_break_depth3():
    for i in range(1):
        try:
            raise ValueError('level1')
        except:
            try:
                raise TypeError('level2')
            except:
                try:
                    raise RuntimeError('level3')
                except:
                    break  # Should clear ALL THREE exceptions
    return 'deep'


assert test_break_depth3() == 'deep', 'break from 3-deep except should reach return'


# Test 3: verify exception stack is empty after break
def test_empty_stack():
    result = []
    for i in range(1):
        try:
            raise ValueError('outer')
        except:
            try:
                raise TypeError('inner')
            except:
                result.append('breaking')
                break
    result.append('after')
    return result


assert test_empty_stack() == ['breaking', 'after'], 'should execute code after break'



# call-external
from pathlib import Path

Path('/nonexistent').read_text()
"""
TRACEBACK:
Traceback (most recent call last):
  File "pathlib__os_read_error.py", line 4, in <module>
    Path('/nonexistent').read_text()
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
FileNotFoundError: [Errno 2] No such file or directory: '/nonexistent'
"""



import string

# === string constants ===
assert string.ascii_lowercase == 'abcdefghijklmnopqrstuvwxyz', 'ascii_lowercase'
assert string.ascii_uppercase == 'ABCDEFGHIJKLMNOPQRSTUVWXYZ', 'ascii_uppercase'
assert string.ascii_letters == 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ', 'ascii_letters'
assert string.digits == '0123456789', 'digits'
assert string.hexdigits == '0123456789abcdefABCDEF', 'hexdigits'
assert string.octdigits == '01234567', 'octdigits'
assert len(string.punctuation) == 32, 'punctuation length'
assert '!' in string.punctuation, 'punctuation has !'
assert '@' in string.punctuation, 'punctuation has @'
assert len(string.whitespace) == 6, 'whitespace length'
assert ' ' in string.whitespace, 'whitespace has space'
assert '\t' in string.whitespace, 'whitespace has tab'
assert '\n' in string.whitespace, 'whitespace has newline'

# === printable contains others ===
for c in string.digits:
    assert c in string.printable, 'printable contains digits'
for c in string.ascii_letters:
    assert c in string.printable, 'printable contains letters'

# === type checks ===
assert isinstance(string.ascii_lowercase, str), 'ascii_lowercase is str'
assert isinstance(string.digits, str), 'digits is str'

# === from import ===
from string import ascii_lowercase, digits

assert ascii_lowercase == 'abcdefghijklmnopqrstuvwxyz', 'from import ascii_lowercase'
assert digits == '0123456789', 'from import digits'



from typing import reveal_type

# === Core types ===
reveal_type(None)
reveal_type(object())
reveal_type(type(1))

# === Primitive types ===
reveal_type(True)
reveal_type(int(1))
reveal_type(float(1.2))

# === String/bytes types ===
reveal_type('hello')
reveal_type(b'foobar')

# === Container types ===
reveal_type([1])
reveal_type((1, 2))
reveal_type({1: 2})
reveal_type({1, 2})
reveal_type(frozenset({1, 2}))
reveal_type(range(10))

# === Iterator types ===
reveal_type(enumerate([1, 2]))
reveal_type(reversed([1, 2]))
reveal_type(zip([1], [2]))

# === Slicing ===
reveal_type(slice(1, 2))

# === Exception types ===
reveal_type(BaseException())
reveal_type(Exception())
reveal_type(SystemExit())
reveal_type(KeyboardInterrupt())
reveal_type(ArithmeticError())
reveal_type(OverflowError())
reveal_type(ZeroDivisionError())
reveal_type(LookupError())
reveal_type(IndexError())
reveal_type(KeyError())
reveal_type(RuntimeError())
reveal_type(NotImplementedError())
reveal_type(RecursionError())
reveal_type(AttributeError())
reveal_type(AssertionError())
reveal_type(MemoryError())
reveal_type(NameError())
reveal_type(SyntaxError())
reveal_type(OSError())
reveal_type(TimeoutError())
reveal_type(TypeError())
reveal_type(ValueError())
reveal_type(StopIteration())



import itertools
import operator

# === chain ===
result = list(itertools.chain([1, 2], [3, 4], [5]))
assert result == [1, 2, 3, 4, 5], 'chain'

result = list(itertools.chain([], [1], []))
assert result == [1], 'chain empty'

# === tee ===
a, b = itertools.tee([1, 2, 3])
assert list(a) == [1, 2, 3], 'tee clone 1'
assert list(b) == [1, 2, 3], 'tee clone 2'

# === islice ===
result = list(itertools.islice([0, 1, 2, 3, 4], 3))
assert result == [0, 1, 2], 'islice stop'

result = list(itertools.islice([0, 1, 2, 3, 4], 1, 4))
assert result == [1, 2, 3], 'islice start stop'

# === zip_longest ===
result = list(itertools.zip_longest([1, 2, 3], [4, 5]))
assert len(result) == 3, 'zip_longest length'
assert result[0] == (1, 4), 'zip_longest 0'
assert result[1] == (2, 5), 'zip_longest 1'
assert result[2] == (3, None), 'zip_longest 2'

# === product ===
result = list(itertools.product([1, 2], [3, 4]))
assert len(result) == 4, 'product length'
assert result[0] == (1, 3), 'product 0'
assert result[3] == (2, 4), 'product 3'

# === permutations ===
result = list(itertools.permutations([1, 2, 3], 2))
assert len(result) == 6, 'permutations length'
assert (1, 2) in result, 'perm (1,2)'
assert (2, 1) in result, 'perm (2,1)'

# === combinations ===
result = list(itertools.combinations([1, 2, 3], 2))
assert len(result) == 3, 'combinations length'
assert (1, 2) in result, 'comb (1,2)'
assert (1, 3) in result, 'comb (1,3)'

# === repeat ===
result = list(itertools.repeat('x', 3))
assert result == ['x', 'x', 'x'], 'repeat'

# === count (infinite counter) ===
from itertools import count, islice

assert list(islice(count(10), 5)) == [10, 11, 12, 13, 14], 'count basic'
assert list(islice(count(0, 2), 4)) == [0, 2, 4, 6], 'count with step'
assert list(islice(count(5, -1), 3)) == [5, 4, 3], 'count negative step'
assert list(islice(count(), 3)) == [0, 1, 2], 'count default args'

# === cycle (infinite repeater) ===
from itertools import cycle

assert list(islice(cycle([1, 2, 3]), 7)) == [1, 2, 3, 1, 2, 3, 1], 'cycle basic'
assert list(islice(cycle('ab'), 5)) == ['a', 'b', 'a', 'b', 'a'], 'cycle string'
assert list(islice(cycle([42]), 3)) == [42, 42, 42], 'cycle single'
assert list(islice(cycle([]), 5)) == [], 'cycle empty'

# === accumulate ===
from itertools import accumulate

assert list(accumulate([1, 2, 3, 4])) == [1, 3, 6, 10], 'accumulate default add'
assert list(accumulate([1])) == [1], 'accumulate single'
assert list(accumulate([])) == [], 'accumulate empty'

assert list(accumulate([1, 2, 3, 4], operator.mul)) == [1, 2, 6, 24], 'accumulate mul'

# === starmap ===
from itertools import starmap

assert list(starmap(operator.add, [(1, 2), (3, 4), (5, 6)])) == [3, 7, 11], 'starmap add'
assert list(starmap(operator.mul, [(2, 3), (4, 5)])) == [6, 20], 'starmap mul'
assert list(starmap(operator.add, [])) == [], 'starmap empty'

# === filterfalse ===
from itertools import filterfalse

assert list(filterfalse(None, [0, 1, '', 'a', False, True])) == [0, '', False], 'filterfalse None pred'
assert list(filterfalse(None, [])) == [], 'filterfalse empty'

# === takewhile ===
from itertools import takewhile

assert list(takewhile(operator.truth, [1, 2, 0, 3, 4])) == [1, 2], 'takewhile truth pred'
assert list(takewhile(operator.truth, [0, 1, 2])) == [], 'takewhile first false'
assert list(takewhile(operator.truth, [1, 2, 3])) == [1, 2, 3], 'takewhile all true'
assert list(takewhile(operator.truth, [])) == [], 'takewhile empty'

# === dropwhile ===
from itertools import dropwhile

assert list(dropwhile(operator.truth, [1, 2, 0, 3, 4])) == [0, 3, 4], 'dropwhile truth pred'
assert list(dropwhile(operator.truth, [0, 1, 2])) == [0, 1, 2], 'dropwhile first false'
assert list(dropwhile(operator.truth, [1, 2, 3])) == [], 'dropwhile all true'
assert list(dropwhile(operator.truth, [])) == [], 'dropwhile empty'

# === compress ===
from itertools import compress

assert list(compress('abcdef', [1, 0, 1, 0, 1, 1])) == ['a', 'c', 'e', 'f'], 'compress basic'
assert list(compress([1, 2, 3], [True, False, True])) == [1, 3], 'compress bool selectors'
assert list(compress([], [])) == [], 'compress empty'
assert list(compress([1, 2, 3], [1])) == [1], 'compress short selectors'

# === pairwise ===
from itertools import pairwise

assert list(pairwise([1, 2, 3, 4])) == [(1, 2), (2, 3), (3, 4)], 'pairwise basic'
assert list(pairwise([1, 2])) == [(1, 2)], 'pairwise two items'
assert list(pairwise([1])) == [], 'pairwise single'
assert list(pairwise([])) == [], 'pairwise empty'
assert list(pairwise('abc')) == [('a', 'b'), ('b', 'c')], 'pairwise string'

# === batched ===
from itertools import batched

assert list(batched([1, 2, 3, 4, 5, 6], 2)) == [(1, 2), (3, 4), (5, 6)], 'batched even'
assert list(batched([1, 2, 3, 4, 5], 2)) == [(1, 2), (3, 4), (5,)], 'batched uneven'
assert list(batched([1, 2, 3], 1)) == [(1,), (2,), (3,)], 'batched size 1'
assert list(batched([], 3)) == [], 'batched empty'
assert list(batched([1, 2, 3], 5)) == [(1, 2, 3)], 'batched size > len'

# === groupby ===
# CPython's groupby returns lazy group iterators; Ouros returns (key, list) tuples.
# We consume each group immediately with list() to work on both.
from itertools import groupby

keys = []
groups = []
for k, g in groupby([1, 1, 2, 2, 2, 3]):
    keys.append(k)
    groups.append(list(g))
assert len(keys) == 3, 'groupby count'
assert keys == [1, 2, 3], 'groupby keys'
assert groups[0] == [1, 1], 'groupby group 0'
assert groups[1] == [2, 2, 2], 'groupby group 1'
assert groups[2] == [3], 'groupby group 2'

assert list(groupby([])) == [], 'groupby empty'

single_keys = []
single_groups = []
for k, g in groupby([5]):
    single_keys.append(k)
    single_groups.append(list(g))
assert single_keys == [5], 'groupby single key'
assert single_groups == [[5]], 'groupby single group'

# === combinations_with_replacement ===
from itertools import combinations_with_replacement

result = list(combinations_with_replacement([1, 2, 3], 2))
assert len(result) == 6, 'comb_wr length'
assert (1, 1) in result, 'comb_wr (1,1)'
assert (1, 2) in result, 'comb_wr (1,2)'
assert (1, 3) in result, 'comb_wr (1,3)'
assert (2, 2) in result, 'comb_wr (2,2)'
assert (2, 3) in result, 'comb_wr (2,3)'
assert (3, 3) in result, 'comb_wr (3,3)'

result0 = list(combinations_with_replacement([1, 2], 0))
assert result0 == [()], 'comb_wr r=0'

result_empty = list(combinations_with_replacement([], 2))
assert result_empty == [], 'comb_wr empty iterable'

# === from import ===
from itertools import chain

assert list(chain([1], [2])) == [1, 2], 'from import chain'
assert list(starmap(pow, [(2, 3)])) == [8], 'from import starmap'



# call-external
# BUG: External calls in recursive functions produce wrong results
# Recursion with external calls doesn't compute the correct value


def sum_with_ext(n):
    if n <= 0:
        return 0
    return add_ints(n, sum_with_ext(n - 1))


# sum_with_ext(3) should compute:
#   add_ints(3, sum_with_ext(2))
#   add_ints(3, add_ints(2, sum_with_ext(1)))
#   add_ints(3, add_ints(2, add_ints(1, sum_with_ext(0))))
#   add_ints(3, add_ints(2, add_ints(1, 0)))
#   = 3 + 2 + 1 = 6
result = sum_with_ext(3)
assert result == 6, 'recursive ext call: 1+2+3=6'



# Mutual reference cycle: a contains b, b contains a
# This creates a cycle where:
#   - a has refcount 2 (variable 'a' + being inside b)
#   - b has refcount 2 (variable 'b' + being inside a)
# Without cycle detection, when both variables go out of scope:
#   - a's refcount drops to 1 (still in b)
#   - b's refcount drops to 1 (still in a)
#   - Neither reaches 0, neither is freed (memory leak)
#
# NOTE: We return len(b) instead of b because repr(b) would cause infinite
# recursion / stack overflow (a separate bug - Python handles this by printing [...]
# for cyclic references)
a = []
b = []
a.append(b)
b.append(a)
len(b)
# ref-counts={'a': 2, 'b': 2}



import asyncio

import pytest
from dirty_equals import IsList
from inline_snapshot import snapshot

import ouros
from ouros import run_async


def test_async():
    code = 'await foobar(1, 2)'
    m = ouros.Sandbox(code, external_functions=['foobar'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)
    assert progress.function_name == snapshot('foobar')
    assert progress.args == snapshot((1, 2))
    call_id = progress.call_id
    progress = progress.resume(future=...)
    assert isinstance(progress, ouros.FutureSnapshot)
    assert progress.pending_call_ids == snapshot([call_id])
    progress = progress.resume({call_id: {'return_value': 3}})
    assert isinstance(progress, ouros.Complete)
    assert progress.output == snapshot(3)


def test_asyncio_gather():
    code = """
import asyncio

await asyncio.gather(foo(1), bar(2))
"""
    m = ouros.Sandbox(code, external_functions=['foo', 'bar'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)
    assert progress.function_name == snapshot('foo')
    assert progress.args == snapshot((1,))
    foo_call_ids = progress.call_id

    progress = progress.resume(future=...)
    assert isinstance(progress, ouros.Snapshot)
    assert progress.function_name == snapshot('bar')
    assert progress.args == snapshot((2,))
    bar_call_ids = progress.call_id
    progress = progress.resume(future=...)

    assert isinstance(progress, ouros.FutureSnapshot)
    dump_progress = progress.dump()

    assert progress.pending_call_ids == IsList(foo_call_ids, bar_call_ids, check_order=False)
    progress = progress.resume({foo_call_ids: {'return_value': 3}, bar_call_ids: {'return_value': 4}})
    assert isinstance(progress, ouros.Complete)
    assert progress.output == snapshot([3, 4])

    progress2 = ouros.FutureSnapshot.load(dump_progress)
    assert progress2.pending_call_ids == IsList(foo_call_ids, bar_call_ids, check_order=False)
    progress = progress2.resume({bar_call_ids: {'return_value': 14}, foo_call_ids: {'return_value': 13}})
    assert isinstance(progress, ouros.Complete)
    assert progress.output == snapshot([13, 14])

    progress3 = ouros.FutureSnapshot.load(dump_progress)
    progress = progress3.resume({bar_call_ids: {'return_value': 14}, foo_call_ids: {'future': ...}})
    assert isinstance(progress, ouros.FutureSnapshot)

    assert progress.pending_call_ids == [foo_call_ids]
    progress = progress.resume({foo_call_ids: {'return_value': 144}})
    assert isinstance(progress, ouros.Complete)
    assert progress.output == snapshot([144, 14])


# === Tests for run_async ===


async def test_run_async_sync_function():
    """Test run_async with a basic sync external function."""
    m = ouros.Sandbox('get_value()', external_functions=['get_value'])

    def get_value():
        return 42

    result = await run_async(m, external_functions={'get_value': get_value})
    assert result == snapshot(42)


async def test_run_async_async_function():
    """Test run_async with a basic async external function."""
    m = ouros.Sandbox('await fetch_data()', external_functions=['fetch_data'])

    async def fetch_data():
        await asyncio.sleep(0.001)
        return 'async result'

    result = await run_async(m, external_functions={'fetch_data': fetch_data})
    assert result == snapshot('async result')


async def test_run_async_function_not_found():
    """Test that missing external function raises wrapped error."""
    m = ouros.Sandbox('missing_func()', external_functions=['missing_func'])

    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        await run_async(m, external_functions={})
    inner = exc_info.value.exception()
    assert isinstance(inner, KeyError)
    assert inner.args[0] == snapshot("'Function missing_func not found'")


async def test_run_async_sync_exception():
    """Test that sync function exceptions propagate correctly."""
    m = ouros.Sandbox('fail()', external_functions=['fail'])

    def fail():
        raise ValueError('sync error')

    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        await run_async(m, external_functions={'fail': fail})
    inner = exc_info.value.exception()
    assert isinstance(inner, ValueError)
    assert inner.args[0] == snapshot('sync error')


async def test_run_async_async_exception():
    """Test that async function exceptions propagate correctly."""
    m = ouros.Sandbox('await async_fail()', external_functions=['async_fail'])

    async def async_fail():
        await asyncio.sleep(0.001)
        raise RuntimeError('async error')

    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        await run_async(m, external_functions={'async_fail': async_fail})
    inner = exc_info.value.exception()
    assert isinstance(inner, RuntimeError)
    assert inner.args[0] == snapshot('async error')


async def test_run_async_exception_caught():
    """Test that exceptions caught in try/except don't propagate."""
    code = """
try:
    fail()
except ValueError:
    caught = True
caught
"""
    m = ouros.Sandbox(code, external_functions=['fail'])

    def fail():
        raise ValueError('caught error')

    result = await run_async(m, external_functions={'fail': fail})
    assert result == snapshot(True)


async def test_run_async_async_exception_caught():
    """Test that async external exceptions are catchable inside sandbox try/except."""
    code = """
try:
    result = await fetch('url')
except ValueError as e:
    result = str(e)
result
"""
    m = ouros.Sandbox(code, external_functions=['fetch'])

    async def fetch(url: str) -> str:
        raise ValueError('network error')

    result = await run_async(m, external_functions={'fetch': fetch})
    assert result == snapshot('network error')


async def test_run_async_multiple_async_functions():
    """Test asyncio.gather with multiple async functions."""
    code = """
import asyncio
await asyncio.gather(fetch_a(), fetch_b())
"""
    m = ouros.Sandbox(code, external_functions=['fetch_a', 'fetch_b'])

    async def fetch_a():
        await asyncio.sleep(0.01)
        return 'a'

    async def fetch_b():
        await asyncio.sleep(0.005)
        return 'b'

    result = await run_async(m, external_functions={'fetch_a': fetch_a, 'fetch_b': fetch_b})
    assert result == snapshot(['a', 'b'])


async def test_run_async_mixed_sync_async():
    """Test mix of sync and async external functions."""
    code = """
sync_val = sync_func()
async_val = await async_func()
sync_val + async_val
"""
    m = ouros.Sandbox(code, external_functions=['sync_func', 'async_func'])

    def sync_func():
        return 10

    async def async_func():
        await asyncio.sleep(0.001)
        return 5

    result = await run_async(m, external_functions={'sync_func': sync_func, 'async_func': async_func})
    assert result == snapshot(15)


async def test_run_async_with_inputs():
    """Test run_async with inputs parameter."""
    m = ouros.Sandbox('process(x, y)', inputs=['x', 'y'], external_functions=['process'])

    def process(a: int, b: int) -> int:
        return a * b

    result = await run_async(m, inputs={'x': 6, 'y': 7}, external_functions={'process': process})
    assert result == snapshot(42)


async def test_run_async_with_print_callback():
    """Test run_async with print_callback parameter."""
    output: list[tuple[str, str]] = []

    def callback(stream: str, text: str) -> None:
        output.append((stream, text))

    m = ouros.Sandbox('print("hello from async")')
    result = await run_async(m, print_callback=callback)
    assert result is None
    assert output == snapshot([('stdout', "hello from async\n")])


async def test_run_async_function_returning_none():
    """Test async function that returns None."""
    m = ouros.Sandbox('do_nothing()', external_functions=['do_nothing'])

    def do_nothing():
        return None

    result = await run_async(m, external_functions={'do_nothing': do_nothing})
    assert result is None


async def test_run_async_no_external_calls():
    """Test run_async when code has no external calls."""
    m = ouros.Sandbox('1 + 2 + 3')
    result = await run_async(m)
    assert result == snapshot(6)


# === Tests for run_async with os parameter ===


async def test_run_async_with_os():
    """run_async can use OSAccess for file operations."""
    from ouros import MemoryFile, OSAccess

    fs = OSAccess([MemoryFile('/test.txt', content='hello world')])

    m = ouros.Sandbox(
        """
from pathlib import Path
Path('/test.txt').read_text()
        """,
        external_functions=[],
    )

    result = await run_async(m, os=fs)
    assert result == snapshot('hello world')


async def test_run_async_os_with_external_functions():
    """run_async can combine OSAccess with external functions."""
    from ouros import MemoryFile, OSAccess

    fs = OSAccess([MemoryFile('/data.txt', content='test data')])

    async def process(text: str) -> str:
        return text.upper()

    m = ouros.Sandbox(
        """
from pathlib import Path
content = Path('/data.txt').read_text()
await process(content)
        """,
        external_functions=['process'],
    )

    result = await run_async(
        m,
        external_functions={'process': process},
        os=fs,
    )
    assert result == snapshot('TEST DATA')


async def test_run_async_os_file_not_found():
    """run_async propagates OS errors correctly."""
    from ouros import OSAccess

    fs = OSAccess()

    m = ouros.Sandbox(
        """
from pathlib import Path
Path('/missing.txt').read_text()
        """,
    )

    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        await run_async(m, os=fs)
    assert str(exc_info.value) == snapshot("FileNotFoundError: [Errno 2] No such file or directory: '/missing.txt'")


async def test_run_async_os_not_provided():
    """run_async raises error when OS function called without os handler."""
    m = ouros.Sandbox(
        """
from pathlib import Path
Path('/test.txt').exists()
        """,
    )

    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        await run_async(m)
    inner = exc_info.value.exception()
    assert isinstance(inner, RuntimeError)
    assert 'OS function' in inner.args[0]
    assert 'no os handler provided' in inner.args[0]


async def test_run_async_os_write_and_read():
    """run_async supports both reading and writing files."""
    from ouros import MemoryFile, OSAccess

    fs = OSAccess([MemoryFile('/file.txt', content='original')])

    m = ouros.Sandbox(
        """
from pathlib import Path
p = Path('/file.txt')
p.write_text('updated')
p.read_text()
        """,
    )

    result = await run_async(m, os=fs)
    assert result == snapshot('updated')



# Test that heap values are properly cleaned up when unexpected kwarg error occurs
def f(a, b):
    return a


# The list [1, 2, 3] should be cleaned up when the error occurs
# because 'c' is an unexpected keyword argument
f([1, 2, 3], [4, 5], c=[6, 7])
# Raise=TypeError("f() got an unexpected keyword argument 'c'")



def foo():
    snap.method()


foo()
"""
TRACEBACK:
Traceback (most recent call last):
  File "traceback__undefined_attr_call.py", line 5, in <module>
    foo()
    ~~~~~
  File "traceback__undefined_attr_call.py", line 2, in foo
    snap.method()
    ~~~~
NameError: name 'snap' is not defined
"""



# === id() returns int type ===
assert isinstance(id(None), int), 'id returns int type'
assert isinstance(id([]), int), 'id of list returns int'
assert isinstance(id('hello'), int), 'id of str returns int'
assert isinstance(id(42), int), 'id of int returns int'

# === Identity operator (is) ===
assert (True is True) == True, 'is True'
assert (False is False) == True, 'is False'
assert (None is None) == True, 'is None'
assert (... is ...) == True, 'is Ellipsis'

# === Identity operator (is not) ===
assert (True is not True) == False, 'is not True'
assert (True is not False) == True, 'is not False'

# === Singleton identity ===
assert id(None) == id(None), 'None singleton'
assert id(True) == id(True), 'True singleton'
assert id(False) == id(False), 'False singleton'
assert id(...) == id(...), 'Ellipsis singleton'

# bool and int are distinct (unlike CPython)
assert id(True) != id(1), 'True is not 1'
assert id(False) != id(0), 'False is not 0'

# distinct singletons
assert id(None) != id(True), 'None is not True'
assert id(None) != id(False), 'None is not False'
assert id(None) != id(...), 'None is not Ellipsis'

# === Integer identity ===
assert id(10) != id(20), 'different ints distinct'

# === Float identity ===
assert id(1.0) != id(2.0), 'different floats distinct'

# === List assignment shares identity ===
lst = [1, 2]
ref = lst
assert id(lst) == id(ref), 'list assignment shared'
assert lst is ref, 'list is same'

# === Variable identity is stable ===
lst = [1, 2]
assert id(lst) == id(lst), 'var id stable'

# === List mutation preserves identity ===
a = [1, 2]
b = a
b.append(3)
assert a is b, 'list mutate preserves identity'

# === Mixed types have distinct ids ===
assert id(1) != id('1'), 'int vs str distinct'
assert id([]) != id(()), 'list vs tuple distinct'

# === Multiple refs share id ===
x = [1, 2]
y = x
z = y
assert id(x) == id(y), 'multiple refs share id xy'
assert id(y) == id(z), 'multiple refs share id yz'

# === String assignment shares identity ===
s = 'hello'
r = s
assert id(s) == id(r), 'str assignment shared'

# === Bytes assignment shares identity ===
b = b'hello'
r = b
assert id(b) == id(r), 'bytes assignment shared'

# === Tuple assignment shares identity ===
t = (1, 2)
r = t
assert id(t) == id(r), 'tuple assignment shared'

# === Boolean is tests ===
assert (True is True) == True, 'bool is test'
assert (False is False) == True, 'bool is test 2'

# === Array is test ===
a = [1, 2]
b = a
assert (a is b) == True, 'array is test'
assert (a is [1, 2]) == False, 'array is new literal'

# === None is tests ===
x = None
assert (x is None) == True, 'var is None'
assert (1 is None) == False, 'int is not None'



import uuid

# === uuid4 basic ===
u = uuid.uuid4()
# For CPython compatibility: convert to string if it's a UUID object
u_str = str(u)
assert len(u_str) == 36, 'uuid4 string length is 36'

# === uuid4 format ===
assert u_str[8] == '-', 'uuid4 dash at 8'
assert u_str[13] == '-', 'uuid4 dash at 13'
assert u_str[18] == '-', 'uuid4 dash at 18'
assert u_str[23] == '-', 'uuid4 dash at 23'

# === uuid4 hex chars only ===
hex_chars = '0123456789abcdef-'
for c in u_str:
    assert c in hex_chars, 'uuid4 valid hex char: ' + c

# === uuid4 version digit ===
assert u_str[14] == '4', 'uuid4 version is 4'

# === uuid4 uniqueness ===
u2 = uuid.uuid4()
assert u != u2, 'uuid4 generates unique values'

# === from import ===
from uuid import uuid4

u3 = uuid4()
u3_str = str(u3)
assert len(u3_str) == 36, 'from import uuid4 length'
assert u3_str[14] == '4', 'from import uuid4 version'



# === Swap list elements via unpack assignment to subscript targets ===
arr = [1, 2, 3, 4, 5]
i = 1
j = 3
arr[i], arr[j] = arr[j], arr[i]
assert arr == [1, 4, 3, 2, 5], f'expected [1, 4, 3, 2, 5], got {arr}'

# === Assign multiple dict entries via unpack assignment to subscripts ===
d = {}
d['a'], d['b'] = 1, 2
assert d['a'] == 1, 'expected d["a"] to be 1, got {}'.format(d['a'])
assert d['b'] == 2, 'expected d["b"] to be 2, got {}'.format(d['b'])



class Descriptor:
    def __set_name__(self, owner, name):
        self.owner = owner
        self.name = name

    def __get__(self, obj, objtype=None):
        return self.name


class MyClass:
    x = Descriptor()
    y = Descriptor()


assert MyClass.x == 'x', '__set_name__ should set name'
assert MyClass.y == 'y', '__set_name__ should set name'



(1, 2, 3)[::0]
"""
TRACEBACK:
Traceback (most recent call last):
  File "slice__step_zero_tuple.py", line 1, in <module>
    (1, 2, 3)[::0]
    ~~~~~~~~~~~~~~
ValueError: slice step cannot be zero
"""



def foo():
    a = []
    a[1]


foo()
"""
TRACEBACK:
Traceback (most recent call last):
  File "traceback__index_error.py", line 6, in <module>
    foo()
    ~~~~~
  File "traceback__index_error.py", line 3, in foo
    a[1]
    ~~~~
IndexError: list index out of range
"""



import operator

# === arithmetic ===
assert operator.add(1, 2) == 3, 'add int'
assert operator.add(1.5, 2.5) == 4.0, 'add float'
assert operator.add('hello', ' world') == 'hello world', 'add str'
assert operator.sub(10, 3) == 7, 'sub int'
assert operator.mul(3, 4) == 12, 'mul int'
assert operator.mul('ab', 3) == 'ababab', 'mul str'
assert operator.truediv(10, 4) == 2.5, 'truediv'
assert operator.floordiv(10, 3) == 3, 'floordiv'
assert operator.mod(10, 3) == 1, 'mod'

# === pow ===
assert operator.pow(2, 3) == 8, 'pow int'
assert operator.pow(2, 0) == 1, 'pow zero'
assert operator.pow(5, 2) == 25, 'pow square'

# === unary ===
assert operator.neg(5) == -5, 'neg int'
assert operator.neg(-3.5) == 3.5, 'neg float'
assert operator.abs(-5) == 5, 'abs int'
assert operator.abs(5) == 5, 'abs positive'
assert operator.abs(-3.5) == 3.5, 'abs float'

# === pos ===
assert operator.pos(5) == 5, 'pos int'
assert operator.pos(-3) == -3, 'pos negative int'
assert operator.pos(3.5) == 3.5, 'pos float'
assert operator.pos(True) == 1, 'pos bool true'
assert operator.pos(False) == 0, 'pos bool false'

# === comparison ===
assert operator.eq(1, 1) is True, 'eq true'
assert operator.eq(1, 2) is False, 'eq false'
assert operator.ne(1, 2) is True, 'ne true'
assert operator.lt(1, 2) is True, 'lt true'
assert operator.lt(2, 1) is False, 'lt false'
assert operator.le(1, 1) is True, 'le equal'
assert operator.le(1, 2) is True, 'le less'
assert operator.gt(2, 1) is True, 'gt true'
assert operator.ge(1, 1) is True, 'ge equal'

# === boolean ===
assert operator.not_(True) is False, 'not_ true'
assert operator.not_(False) is True, 'not_ false'
assert operator.not_(0) is True, 'not_ zero'
assert operator.truth(1) is True, 'truth 1'
assert operator.truth(0) is False, 'truth 0'
assert operator.truth('hello') is True, 'truth str'
assert operator.truth('') is False, 'truth empty str'

# === bitwise ===
assert operator.and_(0b1100, 0b1010) == 0b1000, 'and_'
assert operator.or_(0b1100, 0b1010) == 0b1110, 'or_'
assert operator.xor(0b1100, 0b1010) == 0b0110, 'xor'
assert operator.invert(0) == -1, 'invert 0'
assert operator.invert(1) == -2, 'invert 1'
assert operator.invert(-1) == 0, 'invert -1'
assert operator.invert(True) == -2, 'invert bool'
assert operator.lshift(1, 4) == 16, 'lshift'
assert operator.rshift(16, 4) == 1, 'rshift'

# === index ===
assert operator.index(5) == 5, 'index int'

# === getitem ===
assert operator.getitem([1, 2, 3], 1) == 2, 'getitem list'
assert operator.getitem({'a': 1}, 'a') == 1, 'getitem dict'

# === contains ===
assert operator.contains([1, 2, 3], 2) is True, 'contains list true'
assert operator.contains([1, 2, 3], 5) is False, 'contains list false'
assert operator.contains('hello', 'ell') is True, 'contains str true'

# === concat ===
assert operator.concat([1, 2], [3, 4]) == [1, 2, 3, 4], 'concat lists'
assert operator.concat('foo', 'bar') == 'foobar', 'concat strings'

# === countOf ===
assert operator.countOf([1, 2, 2, 3, 2], 2) == 3, 'countOf list'
assert operator.countOf([1, 2, 3], 5) == 0, 'countOf list not found'
assert operator.countOf([], 1) == 0, 'countOf empty list'
assert operator.countOf((1, 2, 2, 3), 2) == 2, 'countOf tuple'

# === indexOf ===
assert operator.indexOf([10, 20, 30], 20) == 1, 'indexOf list'
assert operator.indexOf([10, 20, 30], 10) == 0, 'indexOf list first'
assert operator.indexOf((5, 10, 15), 15) == 2, 'indexOf tuple'

# === in-place operations ===
assert operator.iadd(3, 4) == 7, 'iadd int'
assert operator.iadd('foo', 'bar') == 'foobar', 'iadd str'
assert operator.isub(10, 3) == 7, 'isub int'
assert operator.imul(3, 4) == 12, 'imul int'
assert operator.itruediv(10, 4) == 2.5, 'itruediv'
assert operator.ifloordiv(10, 3) == 3, 'ifloordiv'
assert operator.imod(10, 3) == 1, 'imod'

# === iadd list (in-place extends) ===
x = [1, 2]
result = operator.iadd(x, [3, 4])
assert result == [1, 2, 3, 4], 'iadd list extends'
assert x == [1, 2, 3, 4], 'iadd list modifies in-place'

# === from import ===
from operator import add, eq

assert add(10, 20) == 30, 'from import add'
assert eq(5, 5) is True, 'from import eq'



# === send(None) to start generator (equivalent to __next__()) ===
def gen_send_start():
    yield 1
    yield 2


g = gen_send_start()
assert g.send(None) == 1, 'send(None) starts generator like next()'
assert g.send(None) == 2, 'send(None) resumes generator like next()'


# === send(value) makes yield expression evaluate to sent value ===
def gen_send_value():
    x = yield 1
    yield x + 10


g = gen_send_value()
assert next(g) == 1, 'first next starts generator'
assert g.send(5) == 15, 'send(5) makes yield evaluate to 5'


# === Multiple sends in sequence ===
def gen_multi_send():
    a = yield 'first'
    b = yield 'second'
    c = yield 'third'
    yield (a, b, c)


g = gen_multi_send()
assert next(g) == 'first', 'initial next'
assert g.send(10) == 'second', 'send 10'
assert g.send(20) == 'third', 'send 20'
assert g.send(30) == (10, 20, 30), 'collected sent values'


# === send() on unstarted generator must use None ===
def gen_send_unstarted():
    x = yield 1
    yield x


g = gen_send_unstarted()
send_error = False
try:
    g.send(5)
except TypeError as e:
    send_error = "can't send non-None value to a just-started generator" in str(e)
assert send_error, 'send(non-None) on unstarted generator raises TypeError'


# === send(None) on unstarted generator is OK ===
def gen_send_none_ok():
    yield 42


g = gen_send_none_ok()
assert g.send(None) == 42, 'send(None) on unstarted generator works'


# === send() on exhausted generator raises StopIteration ===
def gen_send_exhausted():
    yield 1


g = gen_send_exhausted()
next(g)
try:
    next(g)
except StopIteration:
    pass
send_stop = False
try:
    g.send(None)
except StopIteration:
    send_stop = True
assert send_stop, 'send() on exhausted generator raises StopIteration'


# === throw(exception) raises at yield point ===
def gen_throw_basic():
    try:
        yield 1
    except ValueError as e:
        yield 'caught: ' + str(e)


g = gen_throw_basic()
assert next(g) == 1, 'advance to yield'
result = g.throw(ValueError('oops'))
assert result == 'caught: oops', 'throw raises at yield, handler catches and yields'


# === throw() with exception instance ===
def gen_throw_instance():
    try:
        yield 1
    except ValueError as e:
        yield str(e)


g = gen_throw_instance()
next(g)
result = g.throw(ValueError('hello'))
assert result == 'hello', 'throw with exception instance'


# === throw() propagates if not caught ===
def gen_throw_uncaught():
    yield 1
    yield 2


g = gen_throw_uncaught()
next(g)
propagated = False
try:
    g.throw(ValueError('uncaught'))
except ValueError as e:
    propagated = str(e) == 'uncaught'
assert propagated, 'throw propagates uncaught exception to caller'


# === throw() on exhausted generator re-raises ===
def gen_throw_exhausted():
    yield 1


g = gen_throw_exhausted()
next(g)
try:
    next(g)
except StopIteration:
    pass
throw_exhausted = False
try:
    g.throw(ValueError('after'))
except ValueError as e:
    throw_exhausted = str(e) == 'after'
assert throw_exhausted, 'throw on exhausted generator raises exception'


# === throw() on unstarted generator raises immediately ===
def gen_throw_unstarted():
    yield 1


g = gen_throw_unstarted()
throw_unstarted = False
try:
    g.throw(ValueError('before start'))
except ValueError as e:
    throw_unstarted = str(e) == 'before start'
assert throw_unstarted, 'throw on unstarted generator raises exception'


# === close() sends GeneratorExit ===
def gen_close_basic():
    yield 1
    yield 2


g = gen_close_basic()
next(g)
g.close()
close_stop = False
try:
    next(g)
except StopIteration:
    close_stop = True
assert close_stop, 'closed generator raises StopIteration on next()'


# === close() on already-closed generator is no-op ===
def gen_close_twice():
    yield 1


g = gen_close_twice()
next(g)
g.close()
g.close()  # should not raise


# === close() on never-started generator is OK ===
def gen_close_unstarted():
    yield 1


g = gen_close_unstarted()
g.close()  # should not raise


# === close() on exhausted generator is no-op ===
def gen_close_exhausted():
    yield 1


g = gen_close_exhausted()
next(g)
try:
    next(g)
except StopIteration:
    pass
g.close()  # should not raise


# === Generator that catches GeneratorExit and yields raises RuntimeError ===
def gen_bad_close():
    try:
        yield 1
    except GeneratorExit:
        yield 2  # BAD: yielding after GeneratorExit


g = gen_bad_close()
next(g)
runtime_error = False
try:
    g.close()
except RuntimeError:
    runtime_error = True
assert runtime_error, 'yielding after GeneratorExit raises RuntimeError'

# === Generator that catches GeneratorExit and returns is OK ===
cleanup_ran = False


def gen_clean_close():
    global cleanup_ran
    try:
        yield 1
    except GeneratorExit:
        cleanup_ran = True
        return


g = gen_clean_close()
next(g)
g.close()
assert cleanup_ran, 'generator can catch GeneratorExit and return cleanly'


# === Generator that catches GeneratorExit and raises other exception ===
def gen_close_other_exc():
    try:
        yield 1
    except GeneratorExit:
        raise ValueError('custom cleanup error')


g = gen_close_other_exc()
next(g)
other_exc = False
try:
    g.close()
except ValueError as e:
    other_exc = str(e) == 'custom cleanup error'
assert other_exc, 'close() propagates non-GeneratorExit exceptions'


# === send() after generator returns raises StopIteration ===
def gen_send_after_return():
    yield 1
    return 'finished'


g = gen_send_after_return()
next(g)
try:
    next(g)
except StopIteration:
    pass
send_after_stop = False
try:
    g.send(42)
except StopIteration:
    send_after_stop = True
assert send_after_stop, 'send after generator returned raises StopIteration'


# === throw() with generator that catches and re-yields ===
def gen_throw_catch_yield():
    while True:
        try:
            x = yield 'ready'
        except ValueError as e:
            yield 'error: ' + str(e)


g = gen_throw_catch_yield()
assert next(g) == 'ready', 'initial yield'
assert g.throw(ValueError('bad')) == 'error: bad', 'throw caught and new yield'
assert next(g) == 'ready', 'generator continues after throw recovery'


# === send() value used in computation ===
def gen_accumulator():
    total = 0
    while True:
        value = yield total
        if value is None:
            return total
        total = total + value


g = gen_accumulator()
assert next(g) == 0, 'initial total is 0'
assert g.send(10) == 10, 'after sending 10'
assert g.send(20) == 30, 'after sending 20'
assert g.send(5) == 35, 'after sending 5'
ret = None
try:
    g.send(None)
except StopIteration as e:
    ret = e.value
assert ret == 35, 'return value via StopIteration'



# === Generator function returns generator object ===
def gen_simple():
    yield 1


g = gen_simple()
assert type(g).__name__ == 'generator', 'generator function returns generator object'


# === next() yields values and then raises StopIteration ===
def gen_multi():
    yield 1
    yield 2
    yield 3


g = gen_multi()
assert next(g) == 1, 'first next() yields first value'
assert next(g) == 2, 'second next() yields second value'
assert next(g) == 3, 'third next() yields third value'

stopped = False
try:
    next(g)
except StopIteration:
    stopped = True
assert stopped, 'generator raises StopIteration after exhaustion'


# === Generator local state survives across yields ===
def gen_state():
    x = 10
    yield x
    x = x + 7
    yield x


g = gen_state()
assert next(g) == 10, 'first yield sees initial local state'
assert next(g) == 17, 'second yield sees updated local state'


# === Regression: for x in gen() consumes all yielded values ===
def gen_for_regression():
    yield 4
    yield 5
    yield 6


for_values = []
for x in gen_for_regression():
    for_values.append(x)
assert for_values == [4, 5, 6], 'for-loop should resume generator across multiple yields'


# === Regression: list(gen()) collects every value ===
def gen_list_regression():
    yield 'left'
    yield 'middle'
    yield 'right'


assert list(gen_list_regression()) == ['left', 'middle', 'right'], 'list(gen()) should collect all yields'


# === Regression: list comprehension resumes generator repeatedly ===
def gen_comp_regression():
    yield 2
    yield 4
    yield 6


assert [x + 1 for x in gen_comp_regression()] == [3, 5, 7], 'list comprehension should consume full generator'


# === Regression: sum(gen()) consumes every yielded int ===
def gen_sum_regression():
    yield 10
    yield 20
    yield 30


assert sum(gen_sum_regression()) == 60, 'sum(gen()) should consume all yielded ints'


# === Regression: nested generator iteration in list comprehension ===
def inner_gen_regression():
    yield 1
    yield 3
    yield 5


def outer_gen_regression():
    inner = inner_gen_regression()
    yield next(inner)
    yield next(inner)
    yield next(inner)


assert [x for x in outer_gen_regression()] == [1, 3, 5], 'nested generator iteration should resume until exhaustion'

# Return=None



# === Class body vars NOT accessible in methods ===
class ScopeTest:
    x = 10

    def method(self):
        # Class body variable x is NOT in method scope
        # Must access via self.x or ScopeTest.x
        try:
            return x  # NameError
        except NameError:
            return 'no x'

    def correct_method(self):
        return self.x  # Correct: access via self

    def class_access(self):
        return ScopeTest.x  # Also correct: via class name


st = ScopeTest()
assert st.method() == 'no x', 'class body var not in method scope'
assert st.correct_method() == 10, 'class var accessible via self'
assert st.class_access() == 10, 'class var accessible via class name'

# === Class body can reference outer closures ===
# TODO: Requires cell variable support for class bodies (Phase 3)
# def outer():
#     y = 20
#     class Inner:
#         z = y  # This DOES capture y from outer
#         def method(self):
#             return y  # This also captures y from outer
#     return Inner
# InnerCls = outer()
# assert InnerCls.z == 20, 'class body captures outer closure'

# === Class body executes at definition time ===
trace = []


class Tracer:
    trace.append('class body')

    def __init__(self):
        trace.append('init')


assert trace == ['class body'], 'class body runs at definition'
Tracer()
assert trace == ['class body', 'init'], 'init runs at instantiation'
Tracer()
assert trace == ['class body', 'init', 'init'], 'init runs each time'


# === Class variables can reference earlier class variables ===
class VarRef:
    a = 1
    b = a + 2
    c = b * 3


assert VarRef.a == 1, 'first class var'
assert VarRef.b == 3, 'second class var references first'
assert VarRef.c == 9, 'third class var references second'


# === Method cannot reference class variables without self ===
class MethodVarRef:
    x = 10

    def broken(self):
        try:
            return x  # NameError
        except NameError:
            return 'error'

    def correct(self):
        return self.x


mvr = MethodVarRef()
assert mvr.broken() == 'error', 'method cannot access class var without self'
assert mvr.correct() == 10, 'method accesses class var via self'

# === Method captures outer function, not class body ===
# TODO: Requires cell variable support for class bodies (Phase 3)
# def maker(x):
#     class Foo:
#         y = x + 10  # Class body CAN capture outer
#     return Foo

# === Nested class is just a class attribute ===
# TODO: Nested class definitions in class bodies not yet supported
# class Outer:
#     outer_var = 'outer'
#     class Inner:
#         inner_var = 'inner'
# assert Outer.Inner.inner_var == 'inner', 'nested class accessible via outer'
# assert Outer.outer_var == 'outer', 'outer var still accessible'

# === Nested class methods don't see outer class vars ===
# TODO: Requires nested class support
# class Container:
#     container_x = 100
#     class Nested:
#         def method(self):
#             try:
#                 return container_x
#             except NameError:
#                 return 'no access'
# assert Container.Nested().method() == 'no access', 'nested class isolated'


# === Instantiation protocol ===
class Simple:
    def __init__(self, x):
        self.x = x


s = Simple(42)
assert s.x == 42, 'instantiation calls __init__'


# === No __init__ with args raises TypeError ===
class NoInit:
    pass


NoInit()  # OK, no args

try:
    NoInit(1, 2, 3)
    assert False, 'should raise TypeError'
except TypeError as e:
    assert 'takes no arguments' in str(e), 'error message format'



import pytest
from inline_snapshot import snapshot

import ouros


def test_single_input():
    m = ouros.Sandbox('x', inputs=['x'])
    assert m.run(inputs={'x': 42}) == snapshot(42)


def test_multiple_inputs():
    m = ouros.Sandbox('x + y + z', inputs=['x', 'y', 'z'])
    assert m.run(inputs={'x': 1, 'y': 2, 'z': 3}) == snapshot(6)


def test_input_used_in_expression():
    m = ouros.Sandbox('x * 2 + y', inputs=['x', 'y'])
    assert m.run(inputs={'x': 5, 'y': 3}) == snapshot(13)


def test_input_string():
    m = ouros.Sandbox('greeting + " " + name', inputs=['greeting', 'name'])
    assert m.run(inputs={'greeting': 'Hello', 'name': 'World'}) == snapshot('Hello World')


def test_input_list():
    m = ouros.Sandbox('data[0] + data[1]', inputs=['data'])
    assert m.run(inputs={'data': [10, 20]}) == snapshot(30)


def test_input_dict():
    m = ouros.Sandbox('config["a"] * config["b"]', inputs=['config'])
    assert m.run(inputs={'config': {'a': 3, 'b': 4}}) == snapshot(12)


def test_missing_input_raises():
    m = ouros.Sandbox('x + y', inputs=['x', 'y'])
    with pytest.raises(KeyError, match="Missing required input: 'y'"):
        m.run(inputs={'x': 1})


def test_all_inputs_missing_raises():
    m = ouros.Sandbox('x', inputs=['x'])
    with pytest.raises(TypeError, match='Missing required inputs'):
        m.run()


def test_no_inputs_declared_but_provided_raises():
    m = ouros.Sandbox('1 + 1')
    with pytest.raises(TypeError, match='No input variables declared but inputs dict was provided'):
        m.run(inputs={'x': 1})
        with pytest.raises(TypeError, match='No input variables declared but inputs dict was provided'):
            m.run(inputs={})


def test_inputs_order_independent():
    m = ouros.Sandbox('a - b', inputs=['a', 'b'])
    # Dict order shouldn't matter
    assert m.run(inputs={'b': 3, 'a': 10}) == snapshot(7)


def test_function_param_shadows_input():
    """Function parameter should shadow script input with the same name."""
    code = """
def foo(x):
    return x + 1

foo(x * 2)
"""
    m = ouros.Sandbox(code, inputs=['x'])
    # x=5, so foo(x * 2) = foo(10), and inside foo, x is 10 (not 5), so returns 11
    assert m.run(inputs={'x': 5}) == snapshot(11)


def test_function_param_shadows_input_multiple_params():
    """Multiple function parameters should all shadow their corresponding inputs."""
    code = """
def add(x, y):
    return x + y

add(x * 10, y * 100)
"""
    m = ouros.Sandbox(code, inputs=['x', 'y'])
    # x=2, y=3, so add(20, 300) should return 320
    assert m.run(inputs={'x': 2, 'y': 3}) == snapshot(320)


def test_input_accessible_outside_shadowing_function():
    """Script input should still be accessible outside the function that shadows it."""
    code = """
def double(x):
    return x * 2

result = double(10) + x
result
"""
    m = ouros.Sandbox(code, inputs=['x'])
    # double(10) = 20, x (input) = 5, so result = 25
    assert m.run(inputs={'x': 5}) == snapshot(25)


def test_function_param_shadows_input_with_default():
    """Function parameter with default should shadow script input when called with arg."""
    code = """
def foo(x=100):
    return x + 1

foo(x * 2)
"""
    m = ouros.Sandbox(code, inputs=['x'])
    # x=5, foo(10), inside foo x=10 (not 5 or 100), returns 11
    assert m.run(inputs={'x': 5}) == snapshot(11)


def test_function_uses_input_directly():
    """Function that doesn't shadow should still access the input."""
    code = """
def foo(y):
    return x + y

foo(10)
"""
    m = ouros.Sandbox(code, inputs=['x'])
    # x=5 (input), foo(10) with y=10, returns x + y = 5 + 10 = 15
    assert m.run(inputs={'x': 5}) == snapshot(15)



from dataclasses import dataclass


@dataclass
class Point:
    x: int
    y: int


p = Point(1, 2)
assert p.x == 1, 'dataclass field access'
assert p.y == 2, 'dataclass field access'
assert repr(p) == 'Point(x=1, y=2)', 'dataclass repr'
assert Point(1, 2) == Point(1, 2), 'dataclass eq'
assert Point(1, 2) != Point(3, 4), 'dataclass neq'
print('dataclass test passed')



# Test AttributeError message for missing attribute on named tuple
import sys

sys.version_info.foobar
"""
TRACEBACK:
Traceback (most recent call last):
  File "namedtuple__missing_attr.py", line 4, in <module>
    sys.version_info.foobar
AttributeError: 'sys.version_info' object has no attribute 'foobar'
"""



# === Basic exception catching ===
caught = False
try:
    raise ValueError('test')
except ValueError:
    caught = True
assert caught, 'should catch ValueError'

# === Exception variable binding ===
msg = None
try:
    raise TypeError('the message')
except TypeError as e:
    msg = repr(e)
# repr(e) returns "TypeError('the message')" - confirms we caught the right exception
assert msg == "TypeError('the message')", 'should capture exception'

# === Multiple handlers - first match wins ===
which = None
try:
    raise TypeError('type error')
except ValueError:
    which = 'value'
except TypeError:
    which = 'type'
except:
    which = 'bare'
assert which == 'type', 'first matching handler should be used'

# === Bare except catches all ===
caught_bare = False
try:
    raise KeyError('key')
except:
    caught_bare = True
assert caught_bare, 'bare except should catch all'

# === Else block runs when no exception ===
else_ran = False
try:
    x = 1
except:
    pass
else:
    else_ran = True
assert else_ran, 'else should run when no exception'

# === Else block does not run when exception occurs ===
else_ran_with_exc = True
try:
    raise ValueError()
except ValueError:
    pass
else:
    else_ran_with_exc = False
assert else_ran_with_exc, 'else should not run when exception occurs'

# === Finally always runs after try ===
finally_ran = False
try:
    x = 1
finally:
    finally_ran = True
assert finally_ran, 'finally should run after try'

# === Finally runs after exception caught ===
finally_after_catch = False
try:
    raise ValueError()
except ValueError:
    pass
finally:
    finally_after_catch = True
assert finally_after_catch, 'finally should run after exception caught'

# === Bare raise re-raises current exception ===
caught_reraised = False
try:
    try:
        raise ValueError('original')
    except ValueError:
        raise  # bare raise
except ValueError as e:
    caught_reraised = repr(e) == "ValueError('original')"
assert caught_reraised, 'bare raise should re-raise original exception'

# === Nested try/except ===
outer_caught = False
inner_caught = False
try:
    try:
        raise ValueError('inner')
    except ValueError:
        inner_caught = True
        raise TypeError('outer')
except TypeError:
    outer_caught = True
assert inner_caught and outer_caught, 'nested exceptions should work'

# === Exception base class matches all ===
caught_by_base = False
try:
    raise KeyError('key')
except Exception:
    caught_by_base = True
assert caught_by_base, 'Exception should catch all exception types'

# === Tuple of exception types ===
caught_tuple = False
try:
    raise TypeError('type')
except (ValueError, TypeError):
    caught_tuple = True
assert caught_tuple, 'tuple of types should match'


# === Return in try with finally ===
def try_return_finally():
    try:
        return 1
    finally:
        pass


assert try_return_finally() == 1, 'return in try should work with finally'


# === Return in finally overrides try return ===
def finally_return_overrides():
    try:
        return 1
    finally:
        return 2  # type: ignore[returnInFinally]


assert finally_return_overrides() == 2, 'finally return should override try return'

# === Exception in handler propagates ===
handler_exc_propagated = False
try:
    try:
        raise ValueError()
    except ValueError:
        raise TypeError('from handler')
except TypeError as e:
    handler_exc_propagated = repr(e) == "TypeError('from handler')"
assert handler_exc_propagated, 'exception in handler should propagate'


# === Return in finally overrides exception from handler ===
def finally_return_overrides_handler_exc():
    try:
        raise TypeError('Error')
    finally:
        return 'finally wins handler'  # type: ignore


assert finally_return_overrides_handler_exc() == 'finally wins handler', (
    'return in finally should override exception from handler'
)


def finally_return_overrides_handler_exc2():
    try:
        try:
            raise ValueError('inner')
        except ValueError:
            raise TypeError('handler failure')
    finally:
        return 'finally wins handler'  # type: ignore


assert finally_return_overrides_handler_exc2() == 'finally wins handler', (
    'return in finally should override exception from handler'
)


# === Return in finally overrides exception from else ===
def finally_return_overrides_else_exc():
    try:
        try:
            pass
        except ValueError:
            pass
        else:
            raise RuntimeError('else failure')
    finally:
        return 'finally wins else'  # type: ignore


assert finally_return_overrides_else_exc() == 'finally wins else', (
    'return in finally should override exception from else block'
)

# === Exception variable is cleared after handler ===
# After except handler, the exception variable is deleted (Python 3 behavior)
e_cleared = False
try:
    try:
        raise ValueError('test')
    except ValueError as e:
        pass
    # e should be undefined here in Python 3, accessing it raises NameError
    _ = e  # This should raise NameError
except NameError:
    e_cleared = True
assert e_cleared, 'exception variable should be deleted after handler'

# === Unhandled exception propagates ===
unhandled_propagated = False
try:
    try:
        raise KeyError('unhandled')
    except ValueError:
        pass  # KeyError doesn't match, should propagate
except KeyError as e:
    unhandled_propagated = repr(e) == "KeyError('unhandled')"
assert unhandled_propagated, 'unhandled exception should propagate to outer try'

# === Finally runs before unhandled exception propagates ===
finally_before_propagate = False
try:
    try:
        raise KeyError('propagate')
    except ValueError:
        pass
    finally:
        finally_before_propagate = True
except KeyError:
    pass
assert finally_before_propagate, 'finally should run before exception propagates'

# === Exception in finally replaces original exception ===
finally_exc_wins = False
try:
    try:
        raise ValueError('original')
    finally:
        raise TypeError('from finally')
except TypeError as e:
    finally_exc_wins = repr(e) == "TypeError('from finally')"
except ValueError:
    finally_exc_wins = False  # Should not reach here
assert finally_exc_wins, 'exception in finally should replace original'

# === Exception in else propagates ===
else_exc_propagated = False
try:
    try:
        pass  # No exception in try
    except:
        pass
    else:
        raise ValueError('from else')
except ValueError as e:
    else_exc_propagated = repr(e) == "ValueError('from else')"
assert else_exc_propagated, 'exception in else should propagate'

# === Finally runs after exception in else ===
finally_after_else_exc = False
try:
    try:
        pass
    except:
        pass
    else:
        raise ValueError('else error')
    finally:
        finally_after_else_exc = True
except ValueError:
    pass
assert finally_after_else_exc, 'finally should run after exception in else'

# === Exception hierarchy: LookupError ===
# LookupError should catch KeyError
caught_key_by_lookup = False
try:
    raise KeyError('key')
except LookupError:
    caught_key_by_lookup = True
assert caught_key_by_lookup, 'LookupError should catch KeyError'

# LookupError should catch IndexError
caught_index_by_lookup = False
try:
    raise IndexError('index')
except LookupError:
    caught_index_by_lookup = True
assert caught_index_by_lookup, 'LookupError should catch IndexError'

# LookupError should NOT catch ValueError
caught_value_by_lookup = False
try:
    try:
        raise ValueError('value')
    except LookupError:
        caught_value_by_lookup = True
except ValueError:
    pass
assert not caught_value_by_lookup, 'LookupError should NOT catch ValueError'

# === Exception hierarchy: ArithmeticError ===
# ArithmeticError should catch ZeroDivisionError
caught_zero_by_arith = False
try:
    raise ZeroDivisionError('zero')
except ArithmeticError:
    caught_zero_by_arith = True
assert caught_zero_by_arith, 'ArithmeticError should catch ZeroDivisionError'

# ArithmeticError should catch OverflowError
caught_overflow_by_arith = False
try:
    raise OverflowError('overflow')
except ArithmeticError:
    caught_overflow_by_arith = True
assert caught_overflow_by_arith, 'ArithmeticError should catch OverflowError'

# === Exception hierarchy: RuntimeError ===
# RuntimeError should catch NotImplementedError
caught_notimpl_by_runtime = False
try:
    raise NotImplementedError('not impl')
except RuntimeError:
    caught_notimpl_by_runtime = True
assert caught_notimpl_by_runtime, 'RuntimeError should catch NotImplementedError'

# RuntimeError should catch RecursionError
caught_recursion_by_runtime = False
try:
    raise RecursionError('recursion')
except RuntimeError:
    caught_recursion_by_runtime = True
assert caught_recursion_by_runtime, 'RuntimeError should catch RecursionError'

# === Exception hierarchy in tuple ===
# Tuple containing base class should catch derived
caught_by_tuple_base = False
try:
    raise KeyError('key')
except (ValueError, LookupError):
    caught_by_tuple_base = True
assert caught_by_tuple_base, 'tuple with LookupError should catch KeyError'

# === isinstance with exception hierarchy ===
try:
    raise KeyError('key')
except KeyError as e:
    assert isinstance(e, KeyError), 'exception should be instance of KeyError'
    assert isinstance(e, LookupError), 'KeyError should be instance of LookupError'
    assert isinstance(e, Exception), 'KeyError should be instance of Exception'
    assert not isinstance(e, ArithmeticError), 'KeyError should not be ArithmeticError'

try:
    raise ZeroDivisionError('zero')
except ZeroDivisionError as e:
    assert isinstance(e, ZeroDivisionError), 'exception should be instance of ZeroDivisionError'
    assert isinstance(e, ArithmeticError), 'ZeroDivisionError should be instance of ArithmeticError'
    assert isinstance(e, Exception), 'ZeroDivisionError should be instance of Exception'
    assert not isinstance(e, LookupError), 'ZeroDivisionError should not be LookupError'

# === Multiple handlers where none match ===
# Exception should propagate when no handler matches
multi_no_match_propagated = False
try:
    try:
        raise MemoryError('out of memory')
    except ValueError:
        pass
    except TypeError:
        pass
    except KeyError:
        pass
except MemoryError as e:
    multi_no_match_propagated = repr(e) == "MemoryError('out of memory')"
assert multi_no_match_propagated, 'exception should propagate when no handler matches'

# === BaseException hierarchy ===
# BaseException should catch all exceptions including Exception subclasses
caught_value_by_base = False
try:
    raise ValueError('value')
except BaseException:
    caught_value_by_base = True
assert caught_value_by_base, 'BaseException should catch ValueError'

caught_key_by_base = False
try:
    raise KeyError('key')
except BaseException:
    caught_key_by_base = True
assert caught_key_by_base, 'BaseException should catch KeyError'

caught_type_by_base = False
try:
    raise TypeError('type')
except BaseException:
    caught_type_by_base = True
assert caught_type_by_base, 'BaseException should catch TypeError'

# BaseException catches KeyboardInterrupt
caught_keyboard_by_base = False
try:
    raise KeyboardInterrupt()
except BaseException:
    caught_keyboard_by_base = True
assert caught_keyboard_by_base, 'BaseException should catch KeyboardInterrupt'

# BaseException catches SystemExit
caught_sysexit_by_base = False
try:
    raise SystemExit()
except BaseException:
    caught_sysexit_by_base = True
assert caught_sysexit_by_base, 'BaseException should catch SystemExit'

# === Exception does NOT catch BaseException direct subclasses ===
# Exception should NOT catch KeyboardInterrupt
caught_keyboard_by_exc = False
try:
    try:
        raise KeyboardInterrupt()
    except Exception:
        caught_keyboard_by_exc = True
except BaseException:
    pass
assert not caught_keyboard_by_exc, 'Exception should NOT catch KeyboardInterrupt'

# Exception should NOT catch SystemExit
caught_sysexit_by_exc = False
try:
    try:
        raise SystemExit()
    except Exception:
        caught_sysexit_by_exc = True
except BaseException:
    pass
assert not caught_sysexit_by_exc, 'Exception should NOT catch SystemExit'

# But Exception SHOULD catch regular exceptions
caught_value_by_exc = False
try:
    raise ValueError('test')
except Exception:
    caught_value_by_exc = True
assert caught_value_by_exc, 'Exception should catch ValueError'

# === isinstance with BaseException ===
try:
    raise ValueError('test')
except ValueError as e:
    assert isinstance(e, BaseException), 'ValueError should be instance of BaseException'

try:
    raise KeyboardInterrupt()
except KeyboardInterrupt as e:
    assert isinstance(e, BaseException), 'KeyboardInterrupt should be instance of BaseException'
    assert not isinstance(e, Exception), 'KeyboardInterrupt should NOT be instance of Exception'

try:
    raise SystemExit()
except SystemExit as e:
    assert isinstance(e, BaseException), 'SystemExit should be instance of BaseException'
    assert not isinstance(e, Exception), 'SystemExit should NOT be instance of Exception'

# === Tuple containing BaseException ===
caught_by_tuple_with_base = False
try:
    raise KeyboardInterrupt()
except (ValueError, BaseException):
    caught_by_tuple_with_base = True
assert caught_by_tuple_with_base, 'tuple with BaseException should catch KeyboardInterrupt'



a, b, c = (1, 2)
"""
TRACEBACK:
Traceback (most recent call last):
  File "unpack__not_enough.py", line 1, in <module>
    a, b, c = (1, 2)
    ~~~~~~~
ValueError: not enough values to unpack (expected 3, got 2)
"""



# call-external
# BUG: Reference counting bug with string augmented assignment and external calls

# String += with external call causes reference counting error
s = 'hello'
s += concat_strings(' ', 'world')
assert s == 'hello world', 'ext call in augmented string concat'



# Tests for abc module stdlib functions and markers

import abc

# === abc.ABC ===
# ABC should exist as a truthy sentinel
assert abc.ABC, 'ABC should be truthy'

# === abc.ABCMeta ===
# ABCMeta should exist as an accessible attribute
meta = abc.ABCMeta
assert meta is not None, 'ABCMeta should exist'


# === abc.abstractmethod ===
def my_method():
    return 42


decorated = abc.abstractmethod(my_method)
assert decorated is my_method, 'abstractmethod should return the function unchanged'

# === abc.abstractclassmethod ===
# Just verify it exists and can be called as a decorator
decorated2 = abc.abstractclassmethod(my_method)
assert decorated2 is not None, 'abstractclassmethod should return a value'

# === abc.abstractstaticmethod ===
decorated3 = abc.abstractstaticmethod(my_method)
assert decorated3 is not None, 'abstractstaticmethod should return a value'

# === abc.abstractproperty ===
decorated4 = abc.abstractproperty(my_method)
assert decorated4 is not None, 'abstractproperty should return a value'



# === String slicing (getitem_slice, lines 66-75) ===
s = 'hello'
assert s[0:3] == 'hel', 'slice start to mid'
assert s[1:4] == 'ell', 'slice mid range'
assert s[:] == 'hello', 'slice full copy'
assert s[::2] == 'hlo', 'slice with step 2'
assert s[::-1] == 'olleh', 'slice reverse'
assert s[1:4:2] == 'el', 'slice with start stop step'
assert s[-3:] == 'llo', 'slice negative start'
assert s[:-2] == 'hel', 'slice negative stop'

# Unicode slicing
u = 'caf\xe9'
assert u[0:3] == 'caf', 'unicode slice first 3 chars'
assert u[3] == '\xe9', 'unicode slice accented char'

# Empty slice
assert 'hello'[2:2] == '', 'empty slice same indices'
assert 'hello'[3:1] == '', 'empty slice reversed indices'

# Step zero should raise ValueError
try:
    'hello'[::0]
    assert False, 'slice step zero should raise ValueError'
except ValueError:
    pass

# === String iadd with InternString (lines 299-301) ===
s = 'hello'
s += 'x'
assert s == 'hellox', 'iadd single char (InternString)'

s = 'abc'
s += ''
assert s == 'abc', 'iadd empty InternString'

# === String iadd non-str (lines 303) ===
# Trying += with non-string should raise TypeError
try:
    s = 'hello'
    s += 5
    assert False, 'iadd int should raise TypeError'
except TypeError:
    pass

# === Unknown string method (lines 315-316, 335-336, 472-473) ===
try:
    'hello'.nonexistent_method()
    assert False, 'unknown method should raise AttributeError'
except AttributeError as e:
    assert 'str' in str(e), 'error should mention str type'

# === Join with non-string items (lines 532-536) ===
try:
    ','.join([1, 2, 3])
    assert False, 'join with ints should raise TypeError'
except TypeError as e:
    assert 'sequence item' in str(e), 'join error should mention sequence item'

try:
    ','.join(['a', 1])
    assert False, 'join with mixed types should raise TypeError'
except TypeError as e:
    assert 'sequence item' in str(e), 'join mixed error mentions sequence item'

# === String repr with special chars (lines 570-573, 583-586) ===
# Backslash in repr
assert repr('a\\b') == "'a\\\\b'", 'repr with backslash'

# Tab in repr
assert repr('a\tb') == "'a\\tb'", 'repr with tab'

# Carriage return in repr
assert repr('a\rb') == "'a\\rb'", 'repr with carriage return'

# String with single quotes uses double quotes in repr
assert repr("it's") == '"it\'s"', 'repr with single quote uses double quotes'

# String with both quotes (lines 570-573 double-quoted branch)
s = 'it\'s a "test"'
r = repr(s)
assert '\\' in r, 'repr with both quotes contains escapes'

# Backslash/tab/cr in double-quoted branch
assert repr("it's\t") == '"it\'s\\t"', 'repr double-quoted with tab'

# === Unicode digit methods (lines 852-888) ===
# Osmanya digits (U+104A0..U+104A9) - testing isdigit on extended Unicode
# These may not render but they exercise the code paths
osmanya = '\U000104a0'
assert osmanya.isdigit() == True, 'isdigit Osmanya digit'
assert osmanya.isdecimal() == True, 'isdecimal Osmanya digit'

# Brahmi digits (U+11066..U+1106F)
brahmi = '\U00011066'
assert brahmi.isdigit() == True, 'isdigit Brahmi digit'

# === Circled/parenthesized digits in isdigit (lines 915-929) ===
# Circled digit 1 (U+2460)
circled_1 = '\u2460'
assert circled_1.isdigit() == True, 'isdigit circled digit 1'

# Superscript 0 (U+2070)
super_0 = '\u2070'
assert super_0.isdigit() == True, 'isdigit superscript 0'

# Subscript 0 (U+2080)
sub_0 = '\u2080'
assert sub_0.isdigit() == True, 'isdigit subscript 0'

# Parenthesized digit 1 (U+2474)
paren_1 = '\u2474'
assert paren_1.isdigit() == True, 'isdigit parenthesized digit 1'

# Period digit 1 (U+2488)
period_1 = '\u2488'
assert period_1.isdigit() == True, 'isdigit period digit 1'

# Double circled digit 1 (U+24F5)
dbl_circled_1 = '\u24f5'
assert dbl_circled_1.isdigit() == True, 'isdigit double circled digit 1'

# Dingbat circled sans-serif digit 1 (U+2780)
dingbat_1 = '\u2780'
assert dingbat_1.isdigit() == True, 'isdigit dingbat circled digit 1'

# Dingbat negative circled digit 1 (U+278A)
neg_dingbat_1 = '\u278a'
assert neg_dingbat_1.isdigit() == True, 'isdigit negative dingbat digit 1'

# === rindex not found (line 1005) ===
try:
    'hello'.rindex('xyz')
    assert False, 'rindex should raise ValueError for not found'
except ValueError:
    pass

# === Too many args to search methods (lines 1084-1094) ===
try:
    'hello'.find('l', 0, 5, 99)
    assert False, 'find with too many args should raise TypeError'
except TypeError:
    pass

try:
    'hello'.rfind('l', 0, 5, 99)
    assert False, 'rfind with too many args should raise TypeError'
except TypeError:
    pass

try:
    'hello'.index('l', 0, 5, 99)
    assert False, 'index with too many args should raise TypeError'
except TypeError:
    pass

try:
    'hello'.count('l', 0, 5, 99)
    assert False, 'count with too many args should raise TypeError'
except TypeError:
    pass

# === Too many args to startswith/endswith (lines 1160-1170) ===
try:
    'hello'.startswith('h', 0, 5, 99)
    assert False, 'startswith with too many args should raise TypeError'
except TypeError:
    pass

try:
    'hello'.endswith('o', 0, 5, 99)
    assert False, 'endswith with too many args should raise TypeError'
except TypeError:
    pass

# === extract_str_or_tuple_of_str edge cases (lines 1231-1233) ===
# Passing non-string/non-tuple to startswith
try:
    'hello'.startswith(123)
    assert False, 'startswith with int should raise TypeError'
except TypeError:
    pass

try:
    'hello'.endswith(123)
    assert False, 'endswith with int should raise TypeError'
except TypeError:
    pass

# === split/rsplit with whitespace maxsplit (lines 1432-1433, 1478-1479) ===
assert '  a  b  c  '.split(None, 1) == ['a', 'b  c  '], 'split whitespace maxsplit 1'
assert '  a  b  c  '.split(None, 2) == ['a', 'b', 'c  '], 'split whitespace maxsplit 2'
assert '  a  b  c  '.rsplit(None, 1) == ['  a  b', 'c'], 'rsplit whitespace maxsplit 1'
assert '  a  b  c  '.rsplit(None, 2) == ['  a', 'b', 'c'], 'rsplit whitespace maxsplit 2'

# Single word with maxsplit
assert '  hello  '.split(None, 5) == ['hello'], 'split whitespace maxsplit > words'
assert '  hello  '.rsplit(None, 5) == ['hello'], 'rsplit whitespace maxsplit > words'

# === split/rsplit too many positional args (lines 1512-1522) ===
try:
    'a b c'.split(',', 1, 99)
    assert False, 'split with too many args should raise TypeError'
except TypeError:
    pass

try:
    'a b c'.rsplit(',', 1, 99)
    assert False, 'rsplit with too many args should raise TypeError'
except TypeError:
    pass

# === split with None sep positional (lines 1528-1530) ===
assert 'a b c'.split(None) == ['a', 'b', 'c'], 'split None sep explicit'
assert 'a b c'.rsplit(None) == ['a', 'b', 'c'], 'rsplit None sep explicit'

# === splitlines too many positional args (lines 1716-1723) ===
try:
    'a\nb'.splitlines(True, False)
    assert False, 'splitlines with too many args should raise TypeError'
except TypeError:
    pass

# === splitlines with None as keepends (truthy check on None, line 1773-1775) ===
assert 'a\nb'.splitlines(None) == ['a', 'b'], 'splitlines None keepends is falsy'
assert 'a\nb'.splitlines(0) == ['a', 'b'], 'splitlines 0 keepends is falsy'
assert 'a\nb'.splitlines(1) == ['a\n', 'b'], 'splitlines 1 keepends is truthy'

# === partition/rpartition empty sep (lines 1809, 1827, 1842) ===
try:
    'hello'.partition('')
    assert False, 'partition empty sep should raise ValueError'
except ValueError:
    pass

try:
    'hello'.rpartition('')
    assert False, 'rpartition empty sep should raise ValueError'
except ValueError:
    pass

# === replace error paths (lines 1880-1901) ===
# replace with no args
try:
    'hello'.replace()
    assert False, 'replace no args should raise TypeError'
except TypeError:
    pass

# replace with one arg
try:
    'hello'.replace('l')
    assert False, 'replace one arg should raise TypeError'
except TypeError:
    pass

# replace with too many args
try:
    'hello'.replace('l', 'L', 1, 99)
    assert False, 'replace too many args should raise TypeError'
except TypeError:
    pass

# === center/ljust/rjust with negative width (line 2060) ===
assert 'hi'.center(-1) == 'hi', 'center negative width returns original'
assert 'hi'.ljust(-1) == 'hi', 'ljust negative width returns original'
assert 'hi'.rjust(-1) == 'hi', 'rjust negative width returns original'

# === center/ljust/rjust too many args (lines 2045-2052) ===
try:
    'hi'.center(10, '-', 'x')
    assert False, 'center too many args should raise TypeError'
except TypeError:
    pass

try:
    'hi'.ljust(10, '-', 'x')
    assert False, 'ljust too many args should raise TypeError'
except TypeError:
    pass

try:
    'hi'.rjust(10, '-', 'x')
    assert False, 'rjust too many args should raise TypeError'
except TypeError:
    pass

# === center/ljust/rjust fillchar not single char (line 2069) ===
try:
    'hi'.center(10, 'ab')
    assert False, 'center multi-char fillchar should raise TypeError'
except TypeError:
    pass

# === zfill with negative width (line 2090) ===
assert '42'.zfill(-1) == '42', 'zfill negative width'

# === encode with unknown encoding (line 2134) ===
try:
    'hello'.encode('totally-fake-encoding-xyz')
    assert False, 'encode fake encoding should raise LookupError'
except LookupError:
    pass

# === encode with valid error handlers (line 2140) ===
# Test that known error handlers are accepted
assert 'hello'.encode('utf-8', 'strict') == b'hello', 'encode strict handler'
assert 'hello'.encode('utf-8', 'ignore') == b'hello', 'encode ignore handler'
assert 'hello'.encode('utf-8', 'replace') == b'hello', 'encode replace handler'
assert 'hello'.encode('utf-8', 'backslashreplace') == b'hello', 'encode backslashreplace handler'

# === Bool truthiness of strings (lines 249-251) ===
assert bool('') == False, 'empty string is falsy'
assert bool('x') == True, 'non-empty string is truthy'
assert bool('hello') == True, 'longer string is truthy'

# === String Deref (lines 203-205) ===
# Tested implicitly via len, but exercise the path explicitly
s = 'hello'
assert len(s) == 5, 'len on heap string exercises Deref'

# === String from conversion (lines 91-93) ===
# String conversion is implicit in many operations
assert str('hello') == 'hello', 'str() of string'

# === String __contains__ ===
assert 'ell' in 'hello', 'contains substring'
assert 'xyz' not in 'hello', 'not contains substring'
assert '' in 'hello', 'contains empty string'

# === Additional split with sep=None kwarg (lines 1568-1570) ===
assert 'a b c'.split(sep=None) == ['a', 'b', 'c'], 'split sep=None kwarg'

# === Splitlines keepends kwarg (lines 1746-1751) ===
# Already tested above but test the double value error path
# This is tested via splitlines(keepends=True) in existing tests

# === Replace with count kwarg edge cases (lines 1922-1943) ===
# replace with count=0 means replace nothing
assert 'aaa'.replace('a', 'b', 0) == 'aaa', 'replace count 0'



# === C3 MRO conflict ===
class A:
    pass


class B(A):
    pass


# This should raise TypeError: A before B contradicts B's MRO (B -> A -> object)
class C(A, B):
    pass


"""
TRACEBACK:
Traceback (most recent call last):
  File "classes__mro_conflict.py", line 11, in <module>
    class C(A, B):
        pass
TypeError: Cannot create a consistent method resolution order (MRO) for bases A, B
"""



print('xxx', **{"foo'": 123})
"""
TRACEBACK:
Traceback (most recent call last):
  File "builtin__print_invalid_kwarg.py", line 1, in <module>
    print('xxx', **{"foo'": 123})
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
TypeError: print() got an unexpected keyword argument 'foo''
"""



lst = [1, 2, 3]
lst['key'] = 'value'
"""
TRACEBACK:
Traceback (most recent call last):
  File "list__setitem_type_error.py", line 2, in <module>
    lst['key'] = 'value'
    ~~~~~~~~~~
TypeError: list indices must be integers or slices, not str
"""



import concurrent
import concurrent.futures
from concurrent import futures


assert concurrent.futures is not None, 'concurrent_package_has_futures'
assert futures is not None, 'from_concurrent_import_futures'


def square(x):
    return x * x


with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
    immediate = executor.submit(len, [1, 2, 3])
    assert immediate.done() is True, 'submit_builtin_done_true'
    assert immediate.cancelled() is False, 'submit_builtin_not_cancelled'
    assert immediate.result() == 3, 'submit_builtin_result'
    assert immediate.exception() is None, 'submit_builtin_exception_none'

    seen = []

    def on_done(fut):
        seen.append(fut.result())

    callback_ret = immediate.add_done_callback(on_done)
    assert callback_ret is None, 'add_done_callback_return_none'
    assert seen == [3], 'add_done_callback_called'

    deferred = executor.submit(square, 6)
    assert deferred.result() == 36, 'submit_def_result'

    mapped = list(executor.map(square, [1, 2, 3, 4]))
    assert mapped == [1, 4, 9, 16], 'executor_map_results'

    fs = [executor.submit(len, [1]), executor.submit(len, [1, 2])]
    done, not_done = concurrent.futures.wait(fs)
    assert len(done) == 2, 'wait_done_len'
    assert len(not_done) == 0, 'wait_not_done_len'

    completed = list(concurrent.futures.as_completed(fs))
    assert len(completed) == 2, 'as_completed_len'

assert concurrent.futures.ProcessPoolExecutor is not None, 'process_pool_alias_available'

manual = concurrent.futures.Future()
assert manual.done() is False, 'manual_future_initial_done'
assert manual.cancelled() is False, 'manual_future_initial_cancelled'
assert manual.cancel() is True, 'manual_future_cancel_true'
assert manual.done() is True, 'manual_future_done_after_cancel'
assert manual.cancelled() is True, 'manual_future_cancelled_after_cancel'



# call-external
# === Basic external function tests ===

# Simple calls
a = add_ints(10, 20)
assert a == 30, 'add_ints basic'

b = add_ints(-5, 15)
assert b == 10, 'add_ints with negative'

s = concat_strings('hello', ' world')
assert s == 'hello world', 'concat_strings basic'

x = return_value(42)
assert x == 42, 'return_value with int'

y = return_value('test')
assert y == 'test', 'return_value with str'

# === Assignment with external calls ===
result = add_ints(100, 200)
assert result == 300, 'assignment from add_ints'

name = concat_strings('foo', 'bar')
assert name == 'foobar', 'assignment from concat_strings'

# === Nested calls ===
nested = add_ints(1, add_ints(2, 3))
assert nested == 6, 'nested add_ints right'

nested2 = add_ints(add_ints(1, 2), 3)
assert nested2 == 6, 'nested add_ints left'

nested3 = add_ints(add_ints(1, 2), add_ints(3, 4))
assert nested3 == 10, 'nested add_ints both'

deep = add_ints(add_ints(add_ints(1, 2), 3), 4)
assert deep == 10, 'deeply nested add_ints'

# === Chained operations ===
chained = add_ints(1, 2) + add_ints(3, 4)
assert chained == 10, 'chained add_ints with +'

chained2 = add_ints(10, 20) - add_ints(5, 10)
assert chained2 == 15, 'chained add_ints with -'

chained3 = add_ints(2, 3) * add_ints(4, 5)
assert chained3 == 45, 'chained add_ints with *'

str_chain = concat_strings('a', 'b') + concat_strings('c', 'd')
assert str_chain == 'abcd', 'chained concat_strings'

# === External calls in assert statements ===
assert add_ints(5, 5) == 10, 'ext call in assert condition'
assert return_value(True), 'ext call returning truthy in assert'
assert concat_strings('x', 'y') == 'xy', 'concat in assert'
assert add_ints(1, add_ints(2, 3)) == 6, 'nested ext call in assert'

# === Mixed with builtins ===
length = len(concat_strings('hello', 'world'))
assert length == 10, 'len of concat result'

items = [add_ints(1, 2), add_ints(3, 4)]
assert items[0] == 3, 'ext call in list literal first'
assert items[1] == 7, 'ext call in list literal second'

# === Multiple external calls in single expression ===

# Two ext calls added together
sum_result = add_ints(1, 2) + add_ints(3, 4)
assert sum_result == 10, 'two ext calls in addition'

# Three ext calls in one expression
triple = add_ints(1, 1) + add_ints(2, 2) + add_ints(3, 3)
assert triple == 12, 'three ext calls in expression'

# Ext calls in multiplication
mul_result = add_ints(2, 3) * add_ints(1, 1)
assert mul_result == 10, 'ext calls in multiplication'

# Ext calls in subtraction
sub_result = add_ints(10, 5) - add_ints(3, 2)
assert sub_result == 10, 'ext calls in subtraction'

# Complex expression with multiple ext calls
complex_expr = (add_ints(1, 2) + add_ints(3, 4)) * add_ints(0, 2)
assert complex_expr == 20, 'complex expr with ext calls'

# String concatenation with multiple ext calls
str_result = concat_strings(return_value('a'), return_value('b')) + concat_strings('c', 'd')
assert str_result == 'abcd', 'multiple ext calls for string concat'

# Comparison with multiple ext calls
cmp_result = add_ints(5, 5) == add_ints(3, 7)
assert cmp_result == True, 'comparison of two ext call results'

# Nested ext calls in expression
nested_expr = add_ints(add_ints(1, 2), add_ints(3, 4))
assert nested_expr == 10, 'nested ext calls in expression'



# === MappingProxy via type.__dict__ (lines 39-41, 52-55, 57) ===
# Access class namespace via __dict__ which returns a mappingproxy
class MyClass:
    x = 10
    y = 'hello'


mp = MyClass.__dict__

# === MappingProxy len (lines 69-71) ===
assert len(mp) >= 2, 'mappingproxy has at least 2 entries'

# === MappingProxy getitem (lines 84-86, 88-97, 99, 101-102, 107) ===
assert mp['x'] == 10, 'mappingproxy getitem for x'
assert mp['y'] == 'hello', 'mappingproxy getitem for y'

# === MappingProxy getitem KeyError (line 107) ===
try:
    mp['nonexistent']
except KeyError:
    assert True, 'missing key raises KeyError'

# === MappingProxy immutability / setitem error (lines 113-125) ===
try:
    mp['x'] = 99
except TypeError as e:
    assert 'mappingproxy' in str(e), 'setitem on mappingproxy raises TypeError'

# === MappingProxy bool (lines 84-86) ===
assert bool(mp) == True, 'non-empty mappingproxy is truthy'

# === MappingProxy repr (lines 88-97, 99, 101-102) ===
r = repr(mp)
assert 'mappingproxy(' in r, 'repr starts with mappingproxy('

# === MappingProxy .get() method (lines 144, 146-153, 156-158, 160-162) ===
assert mp.get('x') == 10, 'get existing key'
assert mp.get('nonexistent') is None, 'get missing key returns None'
assert mp.get('nonexistent', 42) == 42, 'get missing key returns default'

# === MappingProxy .keys() (lines 165-168) ===
keys = mp.keys()
assert 'x' in keys, 'x in keys'
assert 'y' in keys, 'y in keys'

# === MappingProxy .values() (lines 171-174) ===
values = mp.values()
assert 10 in values, '10 in values'
assert 'hello' in values, 'hello in values'

# === MappingProxy .items() (lines 177-182, 184-185) ===
items = mp.items()
assert ('x', 10) in items, '(x, 10) in items'
assert ('y', 'hello') in items, '(y, hello) in items'

# === MappingProxy .copy() (lines 188-191) ===
d = mp.copy()
assert type(d) == dict, 'copy returns a dict'
assert d['x'] == 10, 'copy preserves values'
assert d['y'] == 'hello', 'copy preserves string values'

# === MappingProxy unknown attribute (lines 127-135, 138-140, 142, 193, 195-196) ===
try:
    mp.nonexistent_method()
except AttributeError as e:
    assert True, 'unknown method raises AttributeError'


# === MappingProxy with empty class ===
class EmptyClass:
    pass


emp = EmptyClass.__dict__
r = repr(emp)
assert 'mappingproxy(' in r, 'empty class repr starts correctly'


# === MappingProxy iteration via keys/values/items ===
class Numbers:
    a = 1
    b = 2
    c = 3


nm = Numbers.__dict__
k = nm.keys()
v = nm.values()
i = nm.items()
assert 'a' in k, 'a in number class keys'
assert 1 in v, '1 in number class values'
assert ('a', 1) in i, '(a, 1) in number class items'



from abc import ABC, abstractmethod


class Animal(ABC):
    @abstractmethod
    def speak(self):
        pass


class Dog(Animal):
    def speak(self):
        return 'Woof'


dog = Dog()
assert dog.speak() == 'Woof', 'concrete method works'

try:
    animal = Animal()
    assert False, 'should not be able to instantiate ABC'
except TypeError as e:
    assert 'abstract' in str(e).lower(), f'wrong error: {e}'

print('abc test passed')



# Test that bytes.startswith with str raises TypeError
b'hello'.startswith('h')
"""
TRACEBACK:
Traceback (most recent call last):
  File "bytes__startswith_str_error.py", line 2, in <module>
    b'hello'.startswith('h')
    ~~~~~~~~~~~~~~~~~~~~~~~~
TypeError: startswith first arg must be bytes or a tuple of bytes, not str
"""



float("it's")
"""
TRACEBACK:
Traceback (most recent call last):
  File "type__float_repr_single_quote.py", line 1, in <module>
    float("it's")
    ~~~~~~~~~~~~~
ValueError: could not convert string to float: "it's"
"""



# === dict.clear() ===
d = {'a': 1, 'b': 2}
d.clear()
assert d == {}, 'clear empties the dict'

d = {}
d.clear()
assert d == {}, 'clear on empty dict is no-op'

# === dict.copy() ===
d = {'a': 1, 'b': 2}
copy = d.copy()
assert copy == {'a': 1, 'b': 2}, 'copy creates equal dict'
assert copy is not d, 'copy creates new dict object'
d['c'] = 3
assert 'c' not in copy, 'copy is independent'

d = {}
copy = d.copy()
assert copy == {}, 'copy empty dict'

# === dict.update() ===
d = {'a': 1}
d.update({'b': 2})
assert d == {'a': 1, 'b': 2}, 'update with dict'

d = {'a': 1}
d.update({'a': 10})
assert d == {'a': 10}, 'update overwrites existing key'

d = {'a': 1}
d.update()
assert d == {'a': 1}, 'update with no args is no-op'

d = {}
d.update([('a', 1), ('b', 2)])
assert d == {'a': 1, 'b': 2}, 'update with list of tuples'

# === dict.setdefault() ===
d = {'a': 1}
result = d.setdefault('a', 10)
assert result == 1, 'setdefault returns existing value'
assert d == {'a': 1}, 'setdefault does not overwrite'

d = {'a': 1}
result = d.setdefault('b', 2)
assert result == 2, 'setdefault returns default for new key'
assert d == {'a': 1, 'b': 2}, 'setdefault inserts new key'

d = {'a': 1}
result = d.setdefault('b')
assert result is None, 'setdefault default is None'
assert d == {'a': 1, 'b': None}, 'setdefault inserts None'

# === dict.popitem() ===
d = {'a': 1, 'b': 2}
item = d.popitem()
assert item == ('b', 2), 'popitem returns last inserted item'
assert d == {'a': 1}, 'popitem removes item'

d = {'x': 10}
item = d.popitem()
assert item == ('x', 10), 'popitem on single-item dict'
assert d == {}, 'dict is now empty'

# === dict.fromkeys() ===
d = dict.fromkeys(['a', 'b', 'c'])
assert d == {'a': None, 'b': None, 'c': None}, 'fromkeys with list, default None'

d = dict.fromkeys(['a', 'b'], 0)
assert d == {'a': 0, 'b': 0}, 'fromkeys with default value'

d = dict.fromkeys([])
assert d == {}, 'fromkeys with empty iterable'

d = dict.fromkeys('abc')
assert d == {'a': None, 'b': None, 'c': None}, 'fromkeys with string iterable'

d = dict.fromkeys(range(3), 'x')
assert d == {0: 'x', 1: 'x', 2: 'x'}, 'fromkeys with range iterable'

d = dict.fromkeys((1, 2, 3), [])
assert d[1] is d[2] and d[2] is d[3], 'fromkeys shares same value object for all keys'

# Duplicate keys - later occurrence wins
d = dict.fromkeys(['a', 'b', 'a'], 1)
assert d == {'a': 1, 'b': 1}, 'fromkeys with duplicate keys'
assert list(d.keys()) == ['a', 'b'], 'fromkeys preserves first occurrence order'

# === dict.fromkeys() instance access ===
# fromkeys is a classmethod but should also work on instances
d = {}.fromkeys(['a', 'b'])
assert d == {'a': None, 'b': None}, 'fromkeys on empty dict instance'

d = {'x': 1}.fromkeys(['a', 'b'], 0)
assert d == {'a': 0, 'b': 0}, 'fromkeys on non-empty dict instance (ignores original)'

# === dict.update() with keyword arguments ===
d = {'a': 1}
d.update(b=2)
assert d == {'a': 1, 'b': 2}, 'update with single kwarg'

d = {'a': 1}
d.update(b=2, c=3)
assert d == {'a': 1, 'b': 2, 'c': 3}, 'update with multiple kwargs'

d = {'a': 1}
d.update(a=10)
assert d == {'a': 10}, 'update kwarg overwrites existing key'

d = {}
d.update(a=1, b=2)
assert d == {'a': 1, 'b': 2}, 'update empty dict with kwargs'

# update with both positional dict and kwargs
d = {'a': 1}
d.update({'b': 2}, c=3)
assert d == {'a': 1, 'b': 2, 'c': 3}, 'update with dict and kwargs'

# kwargs overwrite positional dict values
d = {'a': 1}
d.update({'b': 2}, b=20)
assert d == {'a': 1, 'b': 20}, 'update kwargs overwrite positional dict'

# update with iterable and kwargs
d = {}
d.update([('a', 1)], b=2)
assert d == {'a': 1, 'b': 2}, 'update with iterable and kwargs'

# === Error message for unknown classmethod ===
# Error message should say 'dict' not 'type'
try:
    dict.nonexistent()
    assert False, 'should raise AttributeError'
except AttributeError as e:
    msg = str(e)
    assert 'dict' in msg, f'error should mention dict, got: {e}'
    assert 'nonexistent' in msg, f'error should mention method name, got: {e}'

# === dict.update() sequence element error ===
# Invalid sequence elements should raise ValueError
try:
    d = {}
    d.update([('a', 1), 'x', ('c', 3)])  # 'x' at index 1 is not a 2-tuple
    assert False, 'should raise ValueError'
except (ValueError, TypeError) as e:
    msg = str(e)
    # Error message should mention 'length' requirement
    assert 'length' in msg.lower(), f'error should mention length, got: {e}'
    # TODO: CPython includes element index (#N) in error message
    # assert '#1' in msg, 'error should mention element index'



[1, 2, 3].index(4)
"""
TRACEBACK:
Traceback (most recent call last):
  File "list__index_not_found.py", line 1, in <module>
    [1, 2, 3].index(4)
    ~~~~~~~~~~~~~~~~~~
ValueError: list.index(x): x not in list
"""



import math

# === Constants ===
assert isinstance(math.pi, float), 'pi is float'
assert 3.14159 < math.pi < 3.14160, 'pi value'
assert isinstance(math.e, float), 'e is float'
assert 2.71828 < math.e < 2.71829, 'e value'
assert isinstance(math.tau, float), 'tau is float'
assert abs(math.tau - 2 * math.pi) < 1e-10, 'tau == 2*pi'
assert math.inf > 0, 'inf positive'
assert math.inf > 1e308, 'inf greater than large number'

# === Basic functions ===
assert math.ceil(1.2) == 2, 'ceil 1.2'
assert math.ceil(-1.2) == -1, 'ceil -1.2'
assert math.ceil(3) == 3, 'ceil int'
assert math.floor(1.8) == 1, 'floor 1.8'
assert math.floor(-1.8) == -2, 'floor -1.8'
assert math.floor(3) == 3, 'floor int'
assert math.trunc(1.9) == 1, 'trunc 1.9'
assert math.trunc(-1.9) == -1, 'trunc -1.9'
assert math.fabs(-5.0) == 5.0, 'fabs'
assert math.fabs(5.0) == 5.0, 'fabs positive'

# === Powers and logarithms ===
assert math.sqrt(4.0) == 2.0, 'sqrt 4'
assert math.sqrt(2.0) - 1.4142135623730951 < 1e-10, 'sqrt 2'
assert math.pow(2.0, 3.0) == 8.0, 'pow 2^3'
assert math.pow(4.0, 0.5) == 2.0, 'pow sqrt'
assert abs(math.log(math.e) - 1.0) < 1e-10, 'log e'
assert abs(math.log2(8.0) - 3.0) < 1e-10, 'log2 8'
assert abs(math.log10(1000.0) - 3.0) < 1e-10, 'log10 1000'
assert abs(math.exp(1.0) - math.e) < 1e-10, 'exp 1'
assert math.exp(0.0) == 1.0, 'exp 0'

# === Trigonometry ===
assert abs(math.sin(0.0)) < 1e-10, 'sin 0'
assert abs(math.sin(math.pi / 2) - 1.0) < 1e-10, 'sin pi/2'
assert abs(math.cos(0.0) - 1.0) < 1e-10, 'cos 0'
assert abs(math.tan(0.0)) < 1e-10, 'tan 0'
assert abs(math.asin(1.0) - math.pi / 2) < 1e-10, 'asin 1'
assert abs(math.acos(1.0)) < 1e-10, 'acos 1'
assert abs(math.atan(0.0)) < 1e-10, 'atan 0'
assert abs(math.atan2(1.0, 1.0) - math.pi / 4) < 1e-10, 'atan2'

# === Angle conversion ===
assert abs(math.degrees(math.pi) - 180.0) < 1e-10, 'degrees'
assert abs(math.radians(180.0) - math.pi) < 1e-10, 'radians'

# === Predicates ===
assert math.isnan(math.nan), 'isnan nan'
assert not math.isnan(1.0), 'isnan 1'
assert math.isinf(math.inf), 'isinf inf'
assert not math.isinf(1.0), 'isinf 1'
assert math.isfinite(1.0), 'isfinite 1'
assert not math.isfinite(math.inf), 'isfinite inf'
assert not math.isfinite(math.nan), 'isfinite nan'

# === Integer functions ===
assert math.factorial(0) == 1, 'factorial 0'
assert math.factorial(5) == 120, 'factorial 5'
assert math.factorial(10) == 3628800, 'factorial 10'
assert math.gcd(12, 8) == 4, 'gcd'
assert math.gcd(7, 5) == 1, 'gcd coprime'

# === copysign ===
assert math.copysign(1.0, -1.0) == -1.0, 'copysign'
assert math.copysign(-1.0, 1.0) == 1.0, 'copysign positive'

# === from import ===
from math import ceil, pi, sqrt

assert pi == math.pi, 'from import pi'
assert sqrt(9.0) == 3.0, 'from import sqrt'
assert ceil(1.1) == 2, 'from import ceil'



# Test that bytes.decode raises UnicodeDecodeError for invalid UTF-8
# UnicodeDecodeError is a subclass of ValueError, so it should be caught by both

# Test it raises UnicodeDecodeError
raised_decode_error = False
try:
    b'\xff'.decode()
except UnicodeDecodeError:
    raised_decode_error = True
assert raised_decode_error, 'should raise UnicodeDecodeError for invalid UTF-8'

# Test it can be caught by ValueError (since UnicodeDecodeError is a subclass)
caught_by_value_error = False
try:
    b'\x80\x81'.decode()
except ValueError:
    caught_by_value_error = True
assert caught_by_value_error, 'UnicodeDecodeError should be caught by except ValueError'



import re

# === re.sub with function callback ===
def upper_match(match):
    return match.group(0).upper()

result = re.sub(r'\b\w+\b', upper_match, 'hello world')
assert result == 'HELLO WORLD', f'expected HELLO WORLD, got {result}'

# === re.sub with lambda callback ===
result = re.sub(r'(\d+)', lambda m: str(int(m.group(1)) * 2), 'a1b2c3')
assert result == 'a2b4c6', f'expected a2b4c6, got {result}'

# === re.sub callback accessing groups ===
def swap_groups(m):
    return m.group(2) + '=' + m.group(1)

result = re.sub(r'(\w+):(\w+)', swap_groups, 'key:value')
assert result == 'value=key', f'expected value=key, got {result}'

# === re.sub callback with closure ===
env = {'NAME': 'World', 'GREETING': 'Hello'}
def replacer(m):
    key = m.group(1)
    return env.get(key, m.group(0))

result = re.sub(r'\$\{(\w+)\}', replacer, '${GREETING}, ${NAME}!')
assert result == 'Hello, World!', f'expected Hello, World!, got {result}'

# === re.sub callback with count limit ===
result = re.sub(r'\d+', lambda m: 'X', 'a1b2c3', count=2)
assert result == 'aXbXc3', f'expected aXbXc3, got {result}'

# === re.subn with callback ===
result, count = re.subn(r'\d+', lambda m: str(int(m.group(0)) + 10), 'a1b2c3')
assert result == 'a11b12c13', f'expected a11b12c13, got {result}'
assert count == 3, f'expected 3 substitutions, got {count}'

# === Pattern.sub with callback ===
pat = re.compile(r'\d+')
result = pat.sub(lambda m: str(int(m.group(0)) * 3), '1-2-3')
assert result == '3-6-9', f'expected 3-6-9, got {result}'

# === Pattern.subn with callback ===
result, count = pat.subn(lambda m: str(int(m.group(0)) * 3), '1-2-3')
assert result == '3-6-9', f'expected 3-6-9, got {result}'
assert count == 3, f'expected 3 substitutions, got {count}'

# === re.sub callback returning empty string ===
result = re.sub(r'\d', lambda m: '', 'a1b2c3')
assert result == 'abc', f'expected abc, got {result}'

# === re.sub callback with no matches ===
result = re.sub(r'\d', lambda m: 'X', 'abc')
assert result == 'abc', f'expected abc, got {result}'



# === Basic elif chains ===

# if branch taken
x = 0
if True:
    x = 1
elif True:
    x = 2
assert x == 1, 'if branch should be taken when condition is True'

# elif branch taken (first elif)
y = 0
if False:
    y = 1
elif True:
    y = 2
elif True:
    y = 3
assert y == 2, 'first elif should be taken when if is False and elif is True'

# second elif taken
z = 0
if False:
    z = 1
elif False:
    z = 2
elif True:
    z = 3
assert z == 3, 'second elif should be taken'

# else branch taken after if
a = 0
if False:
    a = 1
else:
    a = 2
assert a == 2, 'else should be taken when if is False'

# else branch taken after elif chain
b = 0
if False:
    b = 1
elif False:
    b = 2
elif False:
    b = 3
else:
    b = 4
assert b == 4, 'else should be taken when all conditions are False'

# === Value-based conditions ===

val = 5

c = 0
if val < 3:
    c = 1
elif val < 6:
    c = 2
elif val < 9:
    c = 3
else:
    c = 4
assert c == 2, 'elif condition val < 6 should match for val=5'

val2 = 10
d = 0
if val2 < 3:
    d = 1
elif val2 < 6:
    d = 2
elif val2 < 9:
    d = 3
else:
    d = 4
assert d == 4, 'else should be taken for val2=10'

# === Multiple statements in branches ===

e = 0
f = 0
if False:
    e = 1
    f = 1
elif True:
    e = 2
    f = 2
else:
    e = 3
    f = 3
assert e == 2, 'first statement in elif executed'
assert f == 2, 'second statement in elif executed'

# === Nested if inside elif ===

g = 0
if False:
    g = 1
elif True:
    if True:
        g = 100
    else:
        g = 200
else:
    g = 3
assert g == 100, 'nested if inside elif should work'

# nested if in else
h = 0
if False:
    h = 1
elif False:
    h = 2
else:
    if True:
        h = 300
    else:
        h = 400
assert h == 300, 'nested if inside else should work'

# === Short-circuit evaluation ===

# elif condition not evaluated if earlier branch taken
called = False


def set_called():
    global called
    called = True
    return True


i = 0
if True:
    i = 1
elif set_called():
    i = 2
assert i == 1, 'if branch taken'
assert called == False, 'elif condition should not be evaluated when if branch is taken'

# reset and test elif evaluation
called = False
j = 0
if False:
    j = 1
elif set_called():
    j = 2
assert j == 2, 'elif branch taken'
assert called == True, 'elif condition should be evaluated when if condition is False'

# === Empty body handling (pass) ===

k = 0
if False:
    pass
elif True:
    k = 1
else:
    pass
assert k == 1, 'elif body executes after if with pass'

# === Boolean expression conditions ===

and_result = 0
if False and True:
    and_result = 1
elif True and True:
    and_result = 2
else:
    and_result = 3
assert and_result == 2, 'elif with and condition'

or_result = 0
if False or False:
    or_result = 1
elif False or True:
    or_result = 2
else:
    or_result = 3
assert or_result == 2, 'elif with or condition'

# === Multiple conditions with and ===

n = 5
o = 0
if n > 1 and n < 3:
    o = 1
elif n > 3 and n < 7:
    o = 2
else:
    o = 3
assert o == 2, 'elif with multiple and conditions'

# === Variable assignment in conditions (walrus operator style via temp var) ===

# Test value propagation through elif chain
p = 0
temp = 10
if temp > 20:
    p = 1
elif temp > 5:
    p = 2
elif temp > 0:
    p = 3
else:
    p = 4
assert p == 2, 'second condition matches temp=10'

# === Equality conditions in branches ===

eq_true_branch = 0
if 7 == 7:
    eq_true_branch = 1
else:
    eq_true_branch = 2
assert eq_true_branch == 1, 'int equality true should keep fallthrough branch'

eq_false_branch = 0
if 7 == 8:
    eq_false_branch = 1
else:
    eq_false_branch = 2
assert eq_false_branch == 2, 'int equality false should jump to else branch'

eq_calls = 0


class EqByValue:
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        global eq_calls
        eq_calls += 1
        return self.value == other.value


lhs = EqByValue(10)
rhs = EqByValue(20)
dunder_false_branch = 0
if lhs == rhs:
    dunder_false_branch = 1
else:
    dunder_false_branch = 2
assert dunder_false_branch == 2, '__eq__ false should take else branch'
assert eq_calls == 1, '__eq__ should be called once for false branch'

rhs.value = 10
dunder_true_branch = 0
if lhs == rhs:
    dunder_true_branch = 1
else:
    dunder_true_branch = 2
assert dunder_true_branch == 1, '__eq__ true should take if branch'
assert eq_calls == 2, '__eq__ should be called once per comparison'



# Test that heap values are cleaned up when duplicate kwargs error occurs
def f(a, b):
    return a


# The list value in the dict should be cleaned up when the error occurs
f(a=[1, 2, 3], **{'a': [4, 5, 6]})
# Raise=TypeError("f() got multiple values for keyword argument 'a'")
